@@ -14,34 +14,128 @@
 package config
 
 import (
+	"encoding/json"
 	"errors"
 	"fmt"
 	"io/ioutil"
+	"net/url"
 	"path/filepath"
 	"regexp"
 	"strings"
 	"time"
 
-	"encoding/json"
-
+	commoncfg "github.com/prometheus/common/config"
 	"github.com/prometheus/common/model"
 	"gopkg.in/yaml.v2"
 )
 
-var patAuthLine = regexp.MustCompile(`((?:api_key|service_key|api_url|token|user_key|password|secret):\s+)(".+"|'.+'|[^\s]+)`)
-
 // Secret is a string that must not be revealed on marshaling.
 type Secret string
 
 // MarshalYAML implements the yaml.Marshaler interface.
 func (s Secret) MarshalYAML() (interface{}, error) {
-	return "<hidden>", nil
+	if s != "" {
+		return "<secret>", nil
+	}
+	return nil, nil
+}
+
+// UnmarshalYAML implements the yaml.Unmarshaler interface for Secret.
+func (s *Secret) UnmarshalYAML(unmarshal func(interface{}) error) error {
+	type plain Secret
+	return unmarshal((*plain)(s))
+}
+
+// MarshalJSON implements the json.Marshaler interface.
+func (s Secret) MarshalJSON() ([]byte, error) {
+	return json.Marshal("<secret>")
+}
+
+// URL is a custom type that allows validation at configuration load time.
+type URL struct {
+	*url.URL
+}
+
+// Copy makes a deep-copy of the struct.
+func (u *URL) Copy() *URL {
+	v := *u.URL
+	return &URL{&v}
+}
+
+// MarshalYAML implements the yaml.Marshaler interface for URL.
+func (u URL) MarshalYAML() (interface{}, error) {
+	if u.URL != nil {
+		return u.URL.String(), nil
+	}
+	return nil, nil
+}
+
+// UnmarshalYAML implements the yaml.Unmarshaler interface for URL.
+func (u *URL) UnmarshalYAML(unmarshal func(interface{}) error) error {
+	var s string
+	if err := unmarshal(&s); err != nil {
+		return err
+	}
+	urlp, err := url.Parse(s)
+	if err != nil {
+		return err
+	}
+	u.URL = urlp
+	return nil
+}
+
+// MarshalJSON implements the json.Marshaler interface for URL.
+func (u URL) MarshalJSON() ([]byte, error) {
+	if u.URL != nil {
+		return json.Marshal(u.URL.String())
+	}
+	return nil, nil
+}
+
+// UnmarshalJSON implements the json.Marshaler interface for URL.
+func (u *URL) UnmarshalJSON(data []byte) error {
+	var s string
+	if err := json.Unmarshal(data, &s); err != nil {
+		return err
+	}
+	urlp, err := url.Parse(s)
+	if err != nil {
+		return err
+	}
+	u.URL = urlp
+	return nil
+}
+
+// SecretURL is a URL that must not be revealed on marshaling.
+type SecretURL URL
+
+// MarshalYAML implements the yaml.Marshaler interface for SecretURL.
+func (s SecretURL) MarshalYAML() (interface{}, error) {
+	if s.URL != nil {
+		return "<secret>", nil
+	}
+	return nil, nil
+}
+
+// UnmarshalYAML implements the yaml.Unmarshaler interface for SecretURL.
+func (s *SecretURL) UnmarshalYAML(unmarshal func(interface{}) error) error {
+	return unmarshal((*URL)(s))
+}
+
+// MarshalJSON implements the json.Marshaler interface for SecretURL.
+func (s SecretURL) MarshalJSON() ([]byte, error) {
+	return json.Marshal("<secret>")
+}
+
+// UnmarshalJSON implements the json.Marshaler interface for SecretURL.
+func (s *SecretURL) UnmarshalJSON(data []byte) error {
+	return json.Unmarshal(data, (*URL)(s))
 }
 
 // Load parses the YAML input s into a Config.
 func Load(s string) (*Config, error) {
 	cfg := &Config{}
-	err := yaml.Unmarshal([]byte(s), cfg)
+	err := yaml.UnmarshalStrict([]byte(s), cfg)
 	if err != nil {
 		return nil, err
 	}
@@ -62,18 +156,18 @@ func Load(s string) (*Config, error) {
 }
 
 // LoadFile parses the given YAML file into a Config.
-func LoadFile(filename string) (*Config, error) {
+func LoadFile(filename string) (*Config, []byte, error) {
 	content, err := ioutil.ReadFile(filename)
 	if err != nil {
-		return nil, err
+		return nil, nil, err
 	}
 	cfg, err := Load(string(content))
 	if err != nil {
-		return nil, err
+		return nil, nil, err
 	}
 
 	resolveFilepaths(filepath.Dir(filename), cfg)
-	return cfg, nil
+	return cfg, content, nil
 }
 
 // resolveFilepaths joins all relative paths in a configuration
@@ -99,36 +193,16 @@ type Config struct {
 	Receivers    []*Receiver    `yaml:"receivers,omitempty" json:"receivers,omitempty"`
 	Templates    []string       `yaml:"templates" json:"templates"`
 
-	// Catches all undefined fields and must be empty after parsing.
-	XXX map[string]interface{} `yaml:",inline" json:"-"`
-
 	// original is the input from which the config was parsed.
 	original string
 }
 
-func checkOverflow(m map[string]interface{}, ctx string) error {
-	if len(m) > 0 {
-		var keys []string
-		for k := range m {
-			keys = append(keys, k)
-		}
-		return fmt.Errorf("unknown fields in %s: %s", ctx, strings.Join(keys, ", "))
-	}
-	return nil
-}
-
 func (c Config) String() string {
-	var s string
-	if c.original != "" {
-		s = c.original
-	} else {
-		b, err := yaml.Marshal(c)
-		if err != nil {
-			return fmt.Sprintf("<error creating config string: %s>", err)
-		}
-		s = string(b)
+	b, err := yaml.Marshal(c)
+	if err != nil {
+		return fmt.Sprintf("<error creating config string: %s>", err)
 	}
-	return patAuthLine.ReplaceAllString(s, "${1}<hidden>")
+	return string(b)
 }
 
 // UnmarshalYAML implements the yaml.Unmarshaler interface.
@@ -154,6 +228,11 @@ func (c *Config) UnmarshalYAML(unmarshal func(interface{}) error) error {
 		if _, ok := names[rcv.Name]; ok {
 			return fmt.Errorf("notification config name %q is not unique", rcv.Name)
 		}
+		for _, wh := range rcv.WebhookConfigs {
+			if wh.HTTPConfig == nil {
+				wh.HTTPConfig = c.Global.HTTPConfig
+			}
+		}
 		for _, ec := range rcv.EmailConfigs {
 			if ec.Smarthost == "" {
 				if c.Global.SMTPSmarthost == "" {
@@ -167,6 +246,9 @@ func (c *Config) UnmarshalYAML(unmarshal func(interface{}) error) error {
 				}
 				ec.From = c.Global.SMTPFrom
 			}
+			if ec.Hello == "" {
+				ec.Hello = c.Global.SMTPHello
+			}
 			if ec.AuthUsername == "" {
 				ec.AuthUsername = c.Global.SMTPAuthUsername
 			}
@@ -185,22 +267,28 @@ func (c *Config) UnmarshalYAML(unmarshal func(interface{}) error) error {
 			}
 		}
 		for _, sc := range rcv.SlackConfigs {
-			if sc.APIURL == "" {
-				if c.Global.SlackAPIURL == "" {
+			if sc.HTTPConfig == nil {
+				sc.HTTPConfig = c.Global.HTTPConfig
+			}
+			if sc.APIURL == nil {
+				if c.Global.SlackAPIURL == nil {
 					return fmt.Errorf("no global Slack API URL set")
 				}
 				sc.APIURL = c.Global.SlackAPIURL
 			}
 		}
 		for _, hc := range rcv.HipchatConfigs {
-			if hc.APIURL == "" {
-				if c.Global.HipchatURL == "" {
+			if hc.HTTPConfig == nil {
+				hc.HTTPConfig = c.Global.HTTPConfig
+			}
+			if hc.APIURL == nil {
+				if c.Global.HipchatAPIURL == nil {
 					return fmt.Errorf("no global Hipchat API URL set")
 				}
-				hc.APIURL = c.Global.HipchatURL
+				hc.APIURL = c.Global.HipchatAPIURL
 			}
-			if !strings.HasSuffix(hc.APIURL, "/") {
-				hc.APIURL += "/"
+			if !strings.HasSuffix(hc.APIURL.Path, "/") {
+				hc.APIURL.Path += "/"
 			}
 			if hc.AuthToken == "" {
 				if c.Global.HipchatAuthToken == "" {
@@ -209,34 +297,90 @@ func (c *Config) UnmarshalYAML(unmarshal func(interface{}) error) error {
 				hc.AuthToken = c.Global.HipchatAuthToken
 			}
 		}
+		for _, poc := range rcv.PushoverConfigs {
+			if poc.HTTPConfig == nil {
+				poc.HTTPConfig = c.Global.HTTPConfig
+			}
+		}
 		for _, pdc := range rcv.PagerdutyConfigs {
-			if pdc.URL == "" {
-				if c.Global.PagerdutyURL == "" {
+			if pdc.HTTPConfig == nil {
+				pdc.HTTPConfig = c.Global.HTTPConfig
+			}
+			if pdc.URL == nil {
+				if c.Global.PagerdutyURL == nil {
 					return fmt.Errorf("no global PagerDuty URL set")
 				}
 				pdc.URL = c.Global.PagerdutyURL
 			}
 		}
 		for _, ogc := range rcv.OpsGenieConfigs {
-			if ogc.APIHost == "" {
-				if c.Global.OpsGenieAPIHost == "" {
+			if ogc.HTTPConfig == nil {
+				ogc.HTTPConfig = c.Global.HTTPConfig
+			}
+			if ogc.APIURL == nil {
+				if c.Global.OpsGenieAPIURL == nil {
 					return fmt.Errorf("no global OpsGenie URL set")
 				}
-				ogc.APIHost = c.Global.OpsGenieAPIHost
+				ogc.APIURL = c.Global.OpsGenieAPIURL
 			}
-			if !strings.HasSuffix(ogc.APIHost, "/") {
-				ogc.APIHost += "/"
+			if !strings.HasSuffix(ogc.APIURL.Path, "/") {
+				ogc.APIURL.Path += "/"
+			}
+			if ogc.APIKey == "" {
+				if c.Global.OpsGenieAPIKey == "" {
+					return fmt.Errorf("no global OpsGenie API Key set")
+				}
+				ogc.APIKey = c.Global.OpsGenieAPIKey
+			}
+		}
+		for _, wcc := range rcv.WechatConfigs {
+			if wcc.HTTPConfig == nil {
+				wcc.HTTPConfig = c.Global.HTTPConfig
+			}
+
+			if wcc.APIURL == nil {
+				if c.Global.WeChatAPIURL == nil {
+					return fmt.Errorf("no global Wechat URL set")
+				}
+				wcc.APIURL = c.Global.WeChatAPIURL
+			}
+
+			if wcc.APISecret == "" {
+				if c.Global.WeChatAPISecret == "" {
+					return fmt.Errorf("no global Wechat ApiSecret set")
+				}
+				wcc.APISecret = c.Global.WeChatAPISecret
+			}
+
+			if wcc.CorpID == "" {
+				if c.Global.WeChatAPICorpID == "" {
+					return fmt.Errorf("no global Wechat CorpID set")
+				}
+				wcc.CorpID = c.Global.WeChatAPICorpID
+			}
+
+			if !strings.HasSuffix(wcc.APIURL.Path, "/") {
+				wcc.APIURL.Path += "/"
 			}
 		}
 		for _, voc := range rcv.VictorOpsConfigs {
-			if voc.APIURL == "" {
-				if c.Global.VictorOpsAPIURL == "" {
+			if voc.HTTPConfig == nil {
+				voc.HTTPConfig = c.Global.HTTPConfig
+			}
+			if voc.APIURL == nil {
+				if c.Global.VictorOpsAPIURL == nil {
 					return fmt.Errorf("no global VictorOps URL set")
 				}
 				voc.APIURL = c.Global.VictorOpsAPIURL
 			}
-			if !strings.HasSuffix(voc.APIURL, "/") {
-				voc.APIURL += "/"
+			if !strings.HasSuffix(voc.APIURL.Path, "/") {
+				voc.APIURL.Path += "/"
+			}
+			if voc.APIKey == "" {
+				if c.Global.VictorOpsAPIKey == "" {
+					return fmt.Errorf("no global VictorOps API Key set")
+				}
+				voc.APIKey = c.Global.VictorOpsAPIKey
 			}
 		}
 		names[rcv.Name] = struct{}{}
@@ -245,21 +389,17 @@ func (c *Config) UnmarshalYAML(unmarshal func(interface{}) error) error {
 	// The root route must not have any matchers as it is the fallback node
 	// for all alerts.
 	if c.Route == nil {
-		return fmt.Errorf("No routes provided")
+		return fmt.Errorf("no routes provided")
 	}
 	if len(c.Route.Receiver) == 0 {
-		return fmt.Errorf("Root route must specify a default receiver")
+		return fmt.Errorf("root route must specify a default receiver")
 	}
 	if len(c.Route.Match) > 0 || len(c.Route.MatchRE) > 0 {
-		return fmt.Errorf("Root route must not have any matchers")
+		return fmt.Errorf("root route must not have any matchers")
 	}
 
 	// Validate that all receivers used in the routing tree are defined.
-	if err := checkReceiver(c.Route, names); err != nil {
-		return err
-	}
-
-	return checkOverflow(c.XXX, "config")
+	return checkReceiver(c.Route, names)
 }
 
 // checkReceiver returns an error if a node in the routing tree
@@ -269,7 +409,7 @@ func checkReceiver(r *Route, receivers map[string]struct{}) error {
 		return nil
 	}
 	if _, ok := receivers[r.Receiver]; !ok {
-		return fmt.Errorf("Undefined receiver %q used in route", r.Receiver)
+		return fmt.Errorf("undefined receiver %q used in route", r.Receiver)
 	}
 	for _, sr := range r.Routes {
 		if err := checkReceiver(sr, receivers); err != nil {
@@ -282,12 +422,23 @@ func checkReceiver(r *Route, receivers map[string]struct{}) error {
 // DefaultGlobalConfig provides global default values.
 var DefaultGlobalConfig = GlobalConfig{
 	ResolveTimeout: model.Duration(5 * time.Minute),
+	HTTPConfig:     &commoncfg.HTTPClientConfig{},
 
+	SMTPHello:       "localhost",
 	SMTPRequireTLS:  true,
-	PagerdutyURL:    "https://events.pagerduty.com/generic/2010-04-15/create_event.json",
-	HipchatURL:      "https://api.hipchat.com/",
-	OpsGenieAPIHost: "https://api.opsgenie.com/",
-	VictorOpsAPIURL: "https://alert.victorops.com/integrations/generic/20131114/alert/",
+	PagerdutyURL:    mustParseURL("https://events.pagerduty.com/v2/enqueue"),
+	HipchatAPIURL:   mustParseURL("https://api.hipchat.com/"),
+	OpsGenieAPIURL:  mustParseURL("https://api.opsgenie.com/"),
+	WeChatAPIURL:    mustParseURL("https://qyapi.weixin.qq.com/cgi-bin/"),
+	VictorOpsAPIURL: mustParseURL("https://alert.victorops.com/integrations/generic/20131114/alert/"),
+}
+
+func mustParseURL(s string) *URL {
+	u, err := url.Parse(s)
+	if err != nil {
+		panic(err)
+	}
+	return &URL{u}
 }
 
 // GlobalConfig defines configuration parameters that are valid globally
@@ -297,32 +448,34 @@ type GlobalConfig struct {
 	// if it has not been updated.
 	ResolveTimeout model.Duration `yaml:"resolve_timeout" json:"resolve_timeout"`
 
-	SMTPFrom         string `yaml:"smtp_from" json:"smtp_from"`
-	SMTPSmarthost    string `yaml:"smtp_smarthost" json:"smtp_smarthost"`
-	SMTPAuthUsername string `yaml:"smtp_auth_username" json:"smtp_auth_username"`
-	SMTPAuthPassword Secret `yaml:"smtp_auth_password" json:"smtp_auth_password"`
-	SMTPAuthSecret   Secret `yaml:"smtp_auth_secret" json:"smtp_auth_secret"`
-	SMTPAuthIdentity string `yaml:"smtp_auth_identity" json:"smtp_auth_identity"`
-	SMTPRequireTLS   bool   `yaml:"smtp_require_tls" json:"smtp_require_tls"`
-	SlackAPIURL      Secret `yaml:"slack_api_url" json:"slack_api_url"`
-	PagerdutyURL     string `yaml:"pagerduty_url" json:"pagerduty_url"`
-	HipchatURL       string `yaml:"hipchat_url" json:"hipchat_url"`
-	HipchatAuthToken Secret `yaml:"hipchat_auth_token" json:"hipchat_auth_token"`
-	OpsGenieAPIHost  string `yaml:"opsgenie_api_host" json:"opsgenie_api_host"`
-	VictorOpsAPIURL  string `yaml:"victorops_api_url" json:"victorops_api_url"`
+	HTTPConfig *commoncfg.HTTPClientConfig `yaml:"http_config,omitempty" json:"http_config,omitempty"`
 
-	// Catches all undefined fields and must be empty after parsing.
-	XXX map[string]interface{} `yaml:",inline" json:"-"`
+	SMTPFrom         string     `yaml:"smtp_from,omitempty" json:"smtp_from,omitempty"`
+	SMTPHello        string     `yaml:"smtp_hello,omitempty" json:"smtp_hello,omitempty"`
+	SMTPSmarthost    string     `yaml:"smtp_smarthost,omitempty" json:"smtp_smarthost,omitempty"`
+	SMTPAuthUsername string     `yaml:"smtp_auth_username,omitempty" json:"smtp_auth_username,omitempty"`
+	SMTPAuthPassword Secret     `yaml:"smtp_auth_password,omitempty" json:"smtp_auth_password,omitempty"`
+	SMTPAuthSecret   Secret     `yaml:"smtp_auth_secret,omitempty" json:"smtp_auth_secret,omitempty"`
+	SMTPAuthIdentity string     `yaml:"smtp_auth_identity,omitempty" json:"smtp_auth_identity,omitempty"`
+	SMTPRequireTLS   bool       `yaml:"smtp_require_tls,omitempty" json:"smtp_require_tls,omitempty"`
+	SlackAPIURL      *SecretURL `yaml:"slack_api_url,omitempty" json:"slack_api_url,omitempty"`
+	PagerdutyURL     *URL       `yaml:"pagerduty_url,omitempty" json:"pagerduty_url,omitempty"`
+	HipchatAPIURL    *URL       `yaml:"hipchat_api_url,omitempty" json:"hipchat_api_url,omitempty"`
+	HipchatAuthToken Secret     `yaml:"hipchat_auth_token,omitempty" json:"hipchat_auth_token,omitempty"`
+	OpsGenieAPIURL   *URL       `yaml:"opsgenie_api_url,omitempty" json:"opsgenie_api_url,omitempty"`
+	OpsGenieAPIKey   Secret     `yaml:"opsgenie_api_key,omitempty" json:"opsgenie_api_key,omitempty"`
+	WeChatAPIURL     *URL       `yaml:"wechat_api_url,omitempty" json:"wechat_api_url,omitempty"`
+	WeChatAPISecret  Secret     `yaml:"wechat_api_secret,omitempty" json:"wechat_api_secret,omitempty"`
+	WeChatAPICorpID  string     `yaml:"wechat_api_corp_id,omitempty" json:"wechat_api_corp_id,omitempty"`
+	VictorOpsAPIURL  *URL       `yaml:"victorops_api_url,omitempty" json:"victorops_api_url,omitempty"`
+	VictorOpsAPIKey  Secret     `yaml:"victorops_api_key,omitempty" json:"victorops_api_key,omitempty"`
 }
 
 // UnmarshalYAML implements the yaml.Unmarshaler interface.
 func (c *GlobalConfig) UnmarshalYAML(unmarshal func(interface{}) error) error {
 	*c = DefaultGlobalConfig
 	type plain GlobalConfig
-	if err := unmarshal((*plain)(c)); err != nil {
-		return err
-	}
-	return checkOverflow(c.XXX, "global")
+	return unmarshal((*plain)(c))
 }
 
 // A Route is a node that contains definitions of how to handle alerts.
@@ -338,9 +491,6 @@ type Route struct {
 	GroupWait      *model.Duration `yaml:"group_wait,omitempty" json:"group_wait,omitempty"`
 	GroupInterval  *model.Duration `yaml:"group_interval,omitempty" json:"group_interval,omitempty"`
 	RepeatInterval *model.Duration `yaml:"repeat_interval,omitempty" json:"repeat_interval,omitempty"`
-
-	// Catches all undefined fields and must be empty after parsing.
-	XXX map[string]interface{} `yaml:",inline" json:"-"`
 }
 
 // UnmarshalYAML implements the yaml.Unmarshaler interface.
@@ -371,7 +521,14 @@ func (r *Route) UnmarshalYAML(unmarshal func(interface{}) error) error {
 		groupBy[ln] = struct{}{}
 	}
 
-	return checkOverflow(r.XXX, "route")
+	if r.GroupInterval != nil && time.Duration(*r.GroupInterval) == time.Duration(0) {
+		return fmt.Errorf("group_interval cannot be zero")
+	}
+	if r.RepeatInterval != nil && time.Duration(*r.RepeatInterval) == time.Duration(0) {
+		return fmt.Errorf("repeat_interval cannot be zero")
+	}
+
+	return nil
 }
 
 // InhibitRule defines an inhibition rule that mutes alerts that match the
@@ -380,22 +537,19 @@ func (r *Route) UnmarshalYAML(unmarshal func(interface{}) error) error {
 type InhibitRule struct {
 	// SourceMatch defines a set of labels that have to equal the given
 	// value for source alerts.
-	SourceMatch map[string]string `yaml:"source_match" json:"source_match"`
+	SourceMatch map[string]string `yaml:"source_match,omitempty" json:"source_match,omitempty"`
 	// SourceMatchRE defines pairs like SourceMatch but does regular expression
 	// matching.
-	SourceMatchRE map[string]Regexp `yaml:"source_match_re" json:"source_match_re"`
+	SourceMatchRE map[string]Regexp `yaml:"source_match_re,omitempty" json:"source_match_re,omitempty"`
 	// TargetMatch defines a set of labels that have to equal the given
 	// value for target alerts.
-	TargetMatch map[string]string `yaml:"target_match" json:"target_match"`
+	TargetMatch map[string]string `yaml:"target_match,omitempty" json:"target_match,omitempty"`
 	// TargetMatchRE defines pairs like TargetMatch but does regular expression
 	// matching.
-	TargetMatchRE map[string]Regexp `yaml:"target_match_re" json:"target_match_re"`
+	TargetMatchRE map[string]Regexp `yaml:"target_match_re,omitempty" json:"target_match_re,omitempty"`
 	// A set of labels that must be equal between the source and target alert
 	// for them to be a match.
-	Equal model.LabelNames `yaml:"equal" json:"equal"`
-
-	// Catches all undefined fields and must be empty after parsing.
-	XXX map[string]interface{} `yaml:",inline" json:"-"`
+	Equal model.LabelNames `yaml:"equal,omitempty" json:"equal,omitempty"`
 }
 
 // UnmarshalYAML implements the yaml.Unmarshaler interface.
@@ -429,7 +583,7 @@ func (r *InhibitRule) UnmarshalYAML(unmarshal func(interface{}) error) error {
 		}
 	}
 
-	return checkOverflow(r.XXX, "inhibit rule")
+	return nil
 }
 
 // Receiver configuration provides configuration on how to contact a receiver.
@@ -443,11 +597,9 @@ type Receiver struct {
 	SlackConfigs     []*SlackConfig     `yaml:"slack_configs,omitempty" json:"slack_configs,omitempty"`
 	WebhookConfigs   []*WebhookConfig   `yaml:"webhook_configs,omitempty" json:"webhook_configs,omitempty"`
 	OpsGenieConfigs  []*OpsGenieConfig  `yaml:"opsgenie_configs,omitempty" json:"opsgenie_configs,omitempty"`
+	WechatConfigs    []*WechatConfig    `yaml:"wechat_configs,omitempty" json:"wechat_configs,omitempty"`
 	PushoverConfigs  []*PushoverConfig  `yaml:"pushover_configs,omitempty" json:"pushover_configs,omitempty"`
 	VictorOpsConfigs []*VictorOpsConfig `yaml:"victorops_configs,omitempty" json:"victorops_configs,omitempty"`
-
-	// Catches all undefined fields and must be empty after parsing.
-	XXX map[string]interface{} `yaml:",inline" json:"-"`
 }
 
 // UnmarshalYAML implements the yaml.Unmarshaler interface.
@@ -459,7 +611,7 @@ func (c *Receiver) UnmarshalYAML(unmarshal func(interface{}) error) error {
 	if c.Name == "" {
 		return fmt.Errorf("missing name in receiver")
 	}
-	return checkOverflow(c.XXX, "receiver config")
+	return nil
 }
 
 // Regexp encapsulates a regexp.Regexp and makes it YAML marshalable.
@@ -482,8 +634,8 @@ func (re *Regexp) UnmarshalYAML(unmarshal func(interface{}) error) error {
 }
 
 // MarshalYAML implements the yaml.Marshaler interface.
-func (re *Regexp) MarshalYAML() (interface{}, error) {
-	if re != nil {
+func (re Regexp) MarshalYAML() (interface{}, error) {
+	if re.Regexp != nil {
 		return re.String(), nil
 	}
 	return nil, nil
@@ -495,7 +647,7 @@ func (re *Regexp) UnmarshalJSON(data []byte) error {
 	if err := json.Unmarshal(data, &s); err != nil {
 		return err
 	}
-	regex, err := regexp.Compile(s)
+	regex, err := regexp.Compile("^(?:" + s + ")$")
 	if err != nil {
 		return err
 	}
