@@ -1,3 +1,16 @@
+// Copyright 2018 Prometheus Team
+// Licensed under the Apache License, Version 2.0 (the "License");
+// you may not use this file except in compliance with the License.
+// You may obtain a copy of the License at
+//
+// http://www.apache.org/licenses/LICENSE-2.0
+//
+// Unless required by applicable law or agreed to in writing, software
+// distributed under the License is distributed on an "AS IS" BASIS,
+// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+// See the License for the specific language governing permissions and
+// limitations under the License.
+
 package dispatch
 
 import (
@@ -6,7 +19,8 @@ import (
 	"sync"
 	"time"
 
-	"github.com/prometheus/common/log"
+	"github.com/go-kit/kit/log"
+	"github.com/go-kit/kit/log/level"
 	"github.com/prometheus/common/model"
 	"github.com/prometheus/prometheus/pkg/labels"
 	"golang.org/x/net/context"
@@ -33,7 +47,7 @@ type Dispatcher struct {
 	ctx    context.Context
 	cancel func()
 
-	log log.Logger
+	logger log.Logger
 }
 
 // NewDispatcher returns a new Dispatcher.
@@ -43,6 +57,7 @@ func NewDispatcher(
 	s notify.Stage,
 	mk types.Marker,
 	to func(time.Duration) time.Duration,
+	l log.Logger,
 ) *Dispatcher {
 	disp := &Dispatcher{
 		alerts:  ap,
@@ -50,7 +65,7 @@ func NewDispatcher(
 		route:   r,
 		marker:  mk,
 		timeout: to,
-		log:     log.With("component", "dispatcher"),
+		logger:  log.With(l, "component", "dispatcher"),
 	}
 	return disp
 }
@@ -80,7 +95,9 @@ type AlertBlock struct {
 // annotated with silencing and inhibition info.
 type APIAlert struct {
 	*model.Alert
-	Status types.AlertStatus `json:"status"`
+	Status      types.AlertStatus `json:"status"`
+	Receivers   []string          `json:"receivers"`
+	Fingerprint string            `json:"fingerprint"`
 }
 
 // AlertGroup is a list of alert blocks grouped by the same label set.
@@ -135,8 +152,9 @@ func (d *Dispatcher) Groups(matchers []*labels.Matcher) AlertOverview {
 				}
 				status := d.marker.Status(a.Fingerprint())
 				aa := &APIAlert{
-					Alert:  a,
-					Status: status,
+					Alert:       a,
+					Status:      status,
+					Fingerprint: a.Fingerprint().String(),
 				}
 
 				if !matchesFilterLabels(aa, matchers) {
@@ -175,16 +193,16 @@ func (d *Dispatcher) run(it provider.AlertIterator) {
 			if !ok {
 				// Iterator exhausted for some reason.
 				if err := it.Err(); err != nil {
-					log.Errorf("Error on alert update: %s", err)
+					level.Error(d.logger).Log("msg", "Error on alert update", "err", err)
 				}
 				return
 			}
 
-			d.log.With("alert", alert).Debug("Received alert")
+			level.Debug(d.logger).Log("msg", "Received alert", "alert", alert)
 
 			// Log errors but keep trying.
 			if err := it.Err(); err != nil {
-				log.Errorf("Error on alert update: %s", err)
+				level.Error(d.logger).Log("msg", "Error on alert update", "err", err)
 				continue
 			}
 
@@ -229,36 +247,37 @@ func (d *Dispatcher) Stop() {
 type notifyFunc func(context.Context, ...*types.Alert) bool
 
 // processAlert determines in which aggregation group the alert falls
-// and insert it.
+// and inserts it.
 func (d *Dispatcher) processAlert(alert *types.Alert, route *Route) {
-	group := model.LabelSet{}
+	groupLabels := model.LabelSet{}
 
 	for ln, lv := range alert.Labels {
 		if _, ok := route.RouteOpts.GroupBy[ln]; ok {
-			group[ln] = lv
+			groupLabels[ln] = lv
 		}
 	}
 
-	fp := group.Fingerprint()
+	fp := groupLabels.Fingerprint()
 
 	d.mtx.Lock()
-	groups, ok := d.aggrGroups[route]
+	defer d.mtx.Unlock()
+
+	group, ok := d.aggrGroups[route]
 	if !ok {
-		groups = map[model.Fingerprint]*aggrGroup{}
-		d.aggrGroups[route] = groups
+		group = map[model.Fingerprint]*aggrGroup{}
+		d.aggrGroups[route] = group
 	}
-	d.mtx.Unlock()
 
 	// If the group does not exist, create it.
-	ag, ok := groups[fp]
+	ag, ok := group[fp]
 	if !ok {
-		ag = newAggrGroup(d.ctx, group, route, d.timeout)
-		groups[fp] = ag
+		ag = newAggrGroup(d.ctx, groupLabels, route, d.timeout, d.logger)
+		group[fp] = ag
 
 		go ag.run(func(ctx context.Context, alerts ...*types.Alert) bool {
-			_, _, err := d.stage.Exec(ctx, alerts...)
+			_, _, err := d.stage.Exec(ctx, d.logger, alerts...)
 			if err != nil {
-				log.Errorf("Notify for %d alerts failed: %s", len(alerts), err)
+				level.Error(d.logger).Log("msg", "Notify for alerts failed", "num_alerts", len(alerts), "err", err)
 			}
 			return err == nil
 		})
@@ -273,7 +292,7 @@ func (d *Dispatcher) processAlert(alert *types.Alert, route *Route) {
 type aggrGroup struct {
 	labels   model.LabelSet
 	opts     *RouteOpts
-	log      log.Logger
+	logger   log.Logger
 	routeKey string
 
 	ctx     context.Context
@@ -282,13 +301,13 @@ type aggrGroup struct {
 	next    *time.Timer
 	timeout func(time.Duration) time.Duration
 
-	mtx     sync.RWMutex
-	alerts  map[model.Fingerprint]*types.Alert
-	hasSent bool
+	mtx        sync.RWMutex
+	alerts     map[model.Fingerprint]*types.Alert
+	hasFlushed bool
 }
 
 // newAggrGroup returns a new aggregation group.
-func newAggrGroup(ctx context.Context, labels model.LabelSet, r *Route, to func(time.Duration) time.Duration) *aggrGroup {
+func newAggrGroup(ctx context.Context, labels model.LabelSet, r *Route, to func(time.Duration) time.Duration, logger log.Logger) *aggrGroup {
 	if to == nil {
 		to = func(d time.Duration) time.Duration { return d }
 	}
@@ -301,7 +320,7 @@ func newAggrGroup(ctx context.Context, labels model.LabelSet, r *Route, to func(
 	}
 	ag.ctx, ag.cancel = context.WithCancel(ctx)
 
-	ag.log = log.With("aggrGroup", ag)
+	ag.logger = log.With(logger, "aggrGroup", ag)
 
 	// Set an initial one-time wait before flushing
 	// the first batch of notifications.
@@ -342,7 +361,7 @@ func (ag *aggrGroup) run(nf notifyFunc) {
 	for {
 		select {
 		case now := <-ag.next.C:
-			// Give the notifcations time until the next flush to
+			// Give the notifications time until the next flush to
 			// finish before terminating them.
 			ctx, cancel := context.WithTimeout(ag.ctx, ag.timeout(ag.opts.GroupInterval))
 
@@ -361,6 +380,7 @@ func (ag *aggrGroup) run(nf notifyFunc) {
 			// Wait the configured interval before calling flush again.
 			ag.mtx.Lock()
 			ag.next.Reset(ag.opts.GroupInterval)
+			ag.hasFlushed = true
 			ag.mtx.Unlock()
 
 			ag.flush(func(alerts ...*types.Alert) bool {
@@ -382,8 +402,7 @@ func (ag *aggrGroup) stop() {
 	<-ag.done
 }
 
-// insert inserts the alert into the aggregation group. If the aggregation group
-// is empty afterwards, it returns true.
+// insert inserts the alert into the aggregation group.
 func (ag *aggrGroup) insert(alert *types.Alert) {
 	ag.mtx.Lock()
 	defer ag.mtx.Unlock()
@@ -392,7 +411,7 @@ func (ag *aggrGroup) insert(alert *types.Alert) {
 
 	// Immediately trigger a flush if the wait duration for this
 	// alert is already over.
-	if !ag.hasSent && alert.StartsAt.Add(ag.opts.GroupWait).Before(time.Now()) {
+	if !ag.hasFlushed && alert.StartsAt.Add(ag.opts.GroupWait).Before(time.Now()) {
 		ag.next.Reset(0)
 	}
 }
@@ -413,16 +432,17 @@ func (ag *aggrGroup) flush(notify func(...*types.Alert) bool) {
 
 	var (
 		alerts      = make(map[model.Fingerprint]*types.Alert, len(ag.alerts))
-		alertsSlice = make([]*types.Alert, 0, len(ag.alerts))
+		alertsSlice = make(types.AlertSlice, 0, len(ag.alerts))
 	)
 	for fp, alert := range ag.alerts {
 		alerts[fp] = alert
 		alertsSlice = append(alertsSlice, alert)
 	}
+	sort.Stable(alertsSlice)
 
 	ag.mtx.Unlock()
 
-	ag.log.Debugln("flushing", alertsSlice)
+	level.Debug(ag.logger).Log("msg", "Flushing", "alerts", fmt.Sprintf("%v", alertsSlice))
 
 	if notify(alertsSlice...) {
 		ag.mtx.Lock()
@@ -433,8 +453,6 @@ func (ag *aggrGroup) flush(notify func(...*types.Alert) bool) {
 				delete(ag.alerts, fp)
 			}
 		}
-
-		ag.hasSent = true
 		ag.mtx.Unlock()
 	}
 }
