@@ -1,11 +1,26 @@
+// Copyright 2018 Prometheus Team
+// Licensed under the Apache License, Version 2.0 (the "License");
+// you may not use this file except in compliance with the License.
+// You may obtain a copy of the License at
+//
+// http://www.apache.org/licenses/LICENSE-2.0
+//
+// Unless required by applicable law or agreed to in writing, software
+// distributed under the License is distributed on an "AS IS" BASIS,
+// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+// See the License for the specific language governing permissions and
+// limitations under the License.
+
 package dispatch
 
 import (
 	"reflect"
 	"sort"
+	"sync"
 	"testing"
 	"time"
 
+	"github.com/go-kit/kit/log"
 	"github.com/prometheus/common/model"
 	"github.com/prometheus/prometheus/pkg/labels"
 	"golang.org/x/net/context"
@@ -144,9 +159,10 @@ func TestAggrGroup(t *testing.T) {
 	)
 
 	var (
-		last     = time.Now()
-		current  = time.Now()
-		alertsCh = make(chan types.AlertSlice)
+		last       = time.Now()
+		current    = time.Now()
+		lastCurMtx = &sync.Mutex{}
+		alertsCh   = make(chan types.AlertSlice)
 	)
 
 	ntfy := func(ctx context.Context, alerts ...*types.Alert) bool {
@@ -167,8 +183,10 @@ func TestAggrGroup(t *testing.T) {
 			t.Errorf("wrong repeat interval: %q", ri)
 		}
 
+		lastCurMtx.Lock()
 		last = current
 		current = time.Now()
+		lastCurMtx.Unlock()
 
 		alertsCh <- types.AlertSlice(alerts)
 
@@ -176,7 +194,7 @@ func TestAggrGroup(t *testing.T) {
 	}
 
 	// Test regular situation where we wait for group_wait to send out alerts.
-	ag := newAggrGroup(context.Background(), lset, route, nil)
+	ag := newAggrGroup(context.Background(), lset, route, nil, log.NewNopLogger())
 	go ag.run(ntfy)
 
 	ag.insert(a1)
@@ -187,7 +205,7 @@ func TestAggrGroup(t *testing.T) {
 
 	case batch := <-alertsCh:
 		if s := time.Since(last); s < opts.GroupWait {
-			t.Fatalf("received batch to early after %v", s)
+			t.Fatalf("received batch too early after %v", s)
 		}
 		exp := types.AlertSlice{a1}
 		sort.Sort(batch)
@@ -207,7 +225,7 @@ func TestAggrGroup(t *testing.T) {
 
 		case batch := <-alertsCh:
 			if s := time.Since(last); s < opts.GroupInterval {
-				t.Fatalf("received batch to early after %v", s)
+				t.Fatalf("received batch too early after %v", s)
 			}
 			exp := types.AlertSlice{a1, a3}
 			sort.Sort(batch)
@@ -224,7 +242,7 @@ func TestAggrGroup(t *testing.T) {
 	// immediate flushing.
 	// Finally, set all alerts to be resolved. After successful notify the aggregation group
 	// should empty itself.
-	ag = newAggrGroup(context.Background(), lset, route, nil)
+	ag = newAggrGroup(context.Background(), lset, route, nil, log.NewNopLogger())
 	go ag.run(ntfy)
 
 	ag.insert(a1)
@@ -253,8 +271,11 @@ func TestAggrGroup(t *testing.T) {
 			t.Fatalf("expected new batch after group interval but received none")
 
 		case batch := <-alertsCh:
-			if s := time.Since(last); s < opts.GroupInterval {
-				t.Fatalf("received batch to early after %v", s)
+			lastCurMtx.Lock()
+			s := time.Since(last)
+			lastCurMtx.Unlock()
+			if s < opts.GroupInterval {
+				t.Fatalf("received batch too early after %v", s)
 			}
 			exp := types.AlertSlice{a1, a2, a3}
 			sort.Sort(batch)
@@ -266,9 +287,12 @@ func TestAggrGroup(t *testing.T) {
 	}
 
 	// Resolve all alerts, they should be removed after the next batch was sent.
-	a1.EndsAt = time.Now()
-	a2.EndsAt = time.Now()
-	a3.EndsAt = time.Now()
+	a1r, a2r, a3r := *a1, *a2, *a3
+	resolved := types.AlertSlice{&a1r, &a2r, &a3r}
+	for _, a := range resolved {
+		a.EndsAt = time.Now()
+		ag.insert(a)
+	}
 
 	select {
 	case <-time.After(2 * opts.GroupInterval):
@@ -276,17 +300,16 @@ func TestAggrGroup(t *testing.T) {
 
 	case batch := <-alertsCh:
 		if s := time.Since(last); s < opts.GroupInterval {
-			t.Fatalf("received batch to early after %v", s)
+			t.Fatalf("received batch too early after %v", s)
 		}
-		exp := types.AlertSlice{a1, a2, a3}
 		sort.Sort(batch)
 
-		if !reflect.DeepEqual(batch, exp) {
-			t.Fatalf("expected alerts %v but got %v", exp, batch)
+		if !reflect.DeepEqual(batch, resolved) {
+			t.Fatalf("expected alerts %v but got %v", resolved, batch)
 		}
 
 		if !ag.empty() {
-			t.Fatalf("Expected aggregation group to be empty after resolving alerts")
+			t.Fatalf("Expected aggregation group to be empty after resolving alerts: %v", ag)
 		}
 	}
 
