@@ -14,11 +14,13 @@
 package inhibit
 
 import (
-	"fmt"
+	"context"
 	"sync"
 	"time"
 
-	"github.com/prometheus/common/log"
+	"github.com/go-kit/kit/log"
+	"github.com/go-kit/kit/log/level"
+	"github.com/oklog/oklog/pkg/group"
 	"github.com/prometheus/common/model"
 
 	"github.com/prometheus/alertmanager/config"
@@ -32,16 +34,18 @@ type Inhibitor struct {
 	alerts provider.Alerts
 	rules  []*InhibitRule
 	marker types.Marker
+	logger log.Logger
 
-	mtx   sync.RWMutex
-	stopc chan struct{}
+	mtx    sync.RWMutex
+	cancel func()
 }
 
 // NewInhibitor returns a new Inhibitor.
-func NewInhibitor(ap provider.Alerts, rs []*config.InhibitRule, mk types.Marker) *Inhibitor {
+func NewInhibitor(ap provider.Alerts, rs []*config.InhibitRule, mk types.Marker, logger log.Logger) *Inhibitor {
 	ih := &Inhibitor{
 		alerts: ap,
 		marker: mk,
+		logger: logger,
 	}
 	for _, cr := range rs {
 		r := NewInhibitRule(cr)
@@ -50,45 +54,33 @@ func NewInhibitor(ap provider.Alerts, rs []*config.InhibitRule, mk types.Marker)
 	return ih
 }
 
-func (ih *Inhibitor) runGC() {
+func (ih *Inhibitor) runGC(ctx context.Context) {
 	for {
 		select {
 		case <-time.After(15 * time.Minute):
 			for _, r := range ih.rules {
 				r.gc()
 			}
-		case <-ih.stopc:
+		case <-ctx.Done():
 			return
 		}
 	}
 }
 
-// Run the Inihibitor's background processing.
-func (ih *Inhibitor) Run() {
-	ih.mtx.Lock()
-	ih.stopc = make(chan struct{})
-	ih.mtx.Unlock()
-
-	go ih.runGC()
-
+func (ih *Inhibitor) run(ctx context.Context) {
 	it := ih.alerts.Subscribe()
 	defer it.Close()
 
 	for {
 		select {
-		case <-ih.stopc:
+		case <-ctx.Done():
 			return
 		case a := <-it.Next():
 			if err := it.Err(); err != nil {
-				log.Errorf("Error iterating alerts: %s", err)
+				level.Error(ih.logger).Log("msg", "Error iterating alerts", "err", err)
 				continue
 			}
-			if a.Resolved() {
-				// As alerts can also time out without an update, we never
-				// handle new resolved alerts but invalidate the cache on read.
-				continue
-			}
-			// Populate the inhibition rules' cache.
+			// Update the inhibition rules' cache.
 			for _, r := range ih.rules {
 				if r.SourceMatchers.Match(a.Labels) {
 					r.set(a)
@@ -98,17 +90,47 @@ func (ih *Inhibitor) Run() {
 	}
 }
 
+// Run the Inihibitor's background processing.
+func (ih *Inhibitor) Run() {
+	var (
+		g   group.Group
+		ctx context.Context
+	)
+
+	ih.mtx.Lock()
+	ctx, ih.cancel = context.WithCancel(context.Background())
+	ih.mtx.Unlock()
+	gcCtx, gcCancel := context.WithCancel(ctx)
+	runCtx, runCancel := context.WithCancel(ctx)
+
+	g.Add(func() error {
+		ih.runGC(gcCtx)
+		return nil
+	}, func(err error) {
+		gcCancel()
+	})
+	g.Add(func() error {
+		ih.run(runCtx)
+		return nil
+	}, func(err error) {
+		runCancel()
+	})
+
+	if err := g.Run(); err != nil {
+		level.Warn(ih.logger).Log("msg", "error running inhibitor", "err", err)
+	}
+}
+
 // Stop the Inhibitor's background processing.
 func (ih *Inhibitor) Stop() {
 	if ih == nil {
 		return
 	}
-	ih.mtx.Lock()
-	defer ih.mtx.Unlock()
 
-	if ih.stopc != nil {
-		close(ih.stopc)
-		ih.stopc = nil
+	ih.mtx.RLock()
+	defer ih.mtx.RUnlock()
+	if ih.cancel != nil {
+		ih.cancel()
 	}
 }
 
@@ -117,8 +139,9 @@ func (ih *Inhibitor) Mutes(lset model.LabelSet) bool {
 	fp := lset.Fingerprint()
 
 	for _, r := range ih.rules {
-		if inhibitedByFP, eq := r.hasEqual(lset); r.TargetMatchers.Match(lset) && eq {
-			ih.marker.SetInhibited(fp, fmt.Sprintf("%d", inhibitedByFP))
+		// Only inhibit if target matchers match but source matchers don't.
+		if inhibitedByFP, eq := r.hasEqual(lset); !r.SourceMatchers.Match(lset) && r.TargetMatchers.Match(lset) && eq {
+			ih.marker.SetInhibited(fp, inhibitedByFP.String())
 			return true
 		}
 	}
