@@ -21,13 +21,14 @@ import (
 
 	"github.com/cenkalti/backoff"
 	"github.com/cespare/xxhash"
+	"github.com/go-kit/kit/log"
+	"github.com/go-kit/kit/log/level"
 	"github.com/prometheus/client_golang/prometheus"
-	"github.com/prometheus/common/log"
 	"github.com/prometheus/common/model"
 	"golang.org/x/net/context"
 
+	"github.com/prometheus/alertmanager/cluster"
 	"github.com/prometheus/alertmanager/config"
-	"github.com/prometheus/alertmanager/inhibit"
 	"github.com/prometheus/alertmanager/nflog"
 	"github.com/prometheus/alertmanager/nflog/nflogpb"
 	"github.com/prometheus/alertmanager/silence"
@@ -47,11 +48,51 @@ var (
 		Name:      "notifications_failed_total",
 		Help:      "The total number of failed notifications.",
 	}, []string{"integration"})
+
+	notificationLatencySeconds = prometheus.NewHistogramVec(prometheus.HistogramOpts{
+		Namespace: "alertmanager",
+		Name:      "notification_latency_seconds",
+		Help:      "The latency of notifications in seconds.",
+		Buckets:   []float64{1, 5, 10, 15, 20},
+	}, []string{"integration"})
 )
 
 func init() {
-	prometheus.Register(numNotifications)
-	prometheus.Register(numFailedNotifications)
+	numNotifications.WithLabelValues("email")
+	numNotifications.WithLabelValues("hipchat")
+	numNotifications.WithLabelValues("pagerduty")
+	numNotifications.WithLabelValues("wechat")
+	numNotifications.WithLabelValues("pushover")
+	numNotifications.WithLabelValues("slack")
+	numNotifications.WithLabelValues("opsgenie")
+	numNotifications.WithLabelValues("webhook")
+	numNotifications.WithLabelValues("victorops")
+	numFailedNotifications.WithLabelValues("email")
+	numFailedNotifications.WithLabelValues("hipchat")
+	numFailedNotifications.WithLabelValues("pagerduty")
+	numFailedNotifications.WithLabelValues("wechat")
+	numFailedNotifications.WithLabelValues("pushover")
+	numFailedNotifications.WithLabelValues("slack")
+	numFailedNotifications.WithLabelValues("opsgenie")
+	numFailedNotifications.WithLabelValues("webhook")
+	numFailedNotifications.WithLabelValues("victorops")
+	notificationLatencySeconds.WithLabelValues("email")
+	notificationLatencySeconds.WithLabelValues("hipchat")
+	notificationLatencySeconds.WithLabelValues("pagerduty")
+	notificationLatencySeconds.WithLabelValues("wechat")
+	notificationLatencySeconds.WithLabelValues("pushover")
+	notificationLatencySeconds.WithLabelValues("slack")
+	notificationLatencySeconds.WithLabelValues("opsgenie")
+	notificationLatencySeconds.WithLabelValues("webhook")
+	notificationLatencySeconds.WithLabelValues("victorops")
+
+	prometheus.MustRegister(numNotifications)
+	prometheus.MustRegister(numFailedNotifications)
+	prometheus.MustRegister(notificationLatencySeconds)
+}
+
+type notifierConfig interface {
+	SendResolved() bool
 }
 
 // MinTimeout is the minimum timeout that is set for the context of a call
@@ -121,10 +162,10 @@ func ReceiverName(ctx context.Context) (string, bool) {
 	return v, ok
 }
 
-func receiverName(ctx context.Context) string {
+func receiverName(ctx context.Context, l log.Logger) string {
 	recv, ok := ReceiverName(ctx)
 	if !ok {
-		log.Error("missing receiver")
+		level.Error(l).Log("msg", "Missing receiver")
 	}
 	return recv
 }
@@ -136,10 +177,10 @@ func GroupKey(ctx context.Context) (string, bool) {
 	return v, ok
 }
 
-func groupLabels(ctx context.Context) model.LabelSet {
+func groupLabels(ctx context.Context, l log.Logger) model.LabelSet {
 	groupLabels, ok := GroupLabels(ctx)
 	if !ok {
-		log.Error("missing group labels")
+		level.Error(l).Log("msg", "Missing group labels")
 	}
 	return groupLabels
 }
@@ -174,15 +215,20 @@ func ResolvedAlerts(ctx context.Context) ([]uint64, bool) {
 
 // A Stage processes alerts under the constraints of the given context.
 type Stage interface {
-	Exec(ctx context.Context, alerts ...*types.Alert) (context.Context, []*types.Alert, error)
+	Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error)
 }
 
 // StageFunc wraps a function to represent a Stage.
-type StageFunc func(ctx context.Context, alerts ...*types.Alert) (context.Context, []*types.Alert, error)
+type StageFunc func(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error)
 
 // Exec implements Stage interface.
-func (f StageFunc) Exec(ctx context.Context, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
-	return f(ctx, alerts...)
+func (f StageFunc) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+	return f(ctx, l, alerts...)
+}
+
+type NotificationLog interface {
+	Log(r *nflogpb.Receiver, gkey string, firingAlerts, resolvedAlerts []uint64) error
+	Query(params ...nflog.QueryParam) ([]*nflogpb.Entry, error)
 }
 
 // BuildPipeline builds a map of receivers to Stages.
@@ -190,26 +236,29 @@ func BuildPipeline(
 	confs []*config.Receiver,
 	tmpl *template.Template,
 	wait func() time.Duration,
-	inhibitor *inhibit.Inhibitor,
+	muter types.Muter,
 	silences *silence.Silences,
-	notificationLog nflog.Log,
+	notificationLog NotificationLog,
 	marker types.Marker,
+	peer *cluster.Peer,
+	logger log.Logger,
 ) RoutingStage {
 	rs := RoutingStage{}
 
-	is := NewInhibitStage(inhibitor, marker)
+	ms := NewGossipSettleStage(peer)
+	is := NewInhibitStage(muter)
 	ss := NewSilenceStage(silences, marker)
 
 	for _, rc := range confs {
-		rs[rc.Name] = MultiStage{is, ss, createStage(rc, tmpl, wait, notificationLog)}
+		rs[rc.Name] = MultiStage{ms, is, ss, createStage(rc, tmpl, wait, notificationLog, logger)}
 	}
 	return rs
 }
 
 // createStage creates a pipeline of stages for a receiver.
-func createStage(rc *config.Receiver, tmpl *template.Template, wait func() time.Duration, notificationLog nflog.Log) Stage {
+func createStage(rc *config.Receiver, tmpl *template.Template, wait func() time.Duration, notificationLog NotificationLog, logger log.Logger) Stage {
 	var fs FanoutStage
-	for _, i := range BuildReceiverIntegrations(rc, tmpl) {
+	for _, i := range BuildReceiverIntegrations(rc, tmpl, logger) {
 		recv := &nflogpb.Receiver{
 			GroupName:   rc.Name,
 			Integration: i.name,
@@ -217,8 +266,8 @@ func createStage(rc *config.Receiver, tmpl *template.Template, wait func() time.
 		}
 		var s MultiStage
 		s = append(s, NewWaitStage(wait))
-		s = append(s, NewDedupStage(notificationLog, recv, i.conf.SendResolved()))
-		s = append(s, NewRetryStage(i))
+		s = append(s, NewDedupStage(i, notificationLog, recv))
+		s = append(s, NewRetryStage(i, rc.Name))
 		s = append(s, NewSetNotifiesStage(notificationLog, recv))
 
 		fs = append(fs, s)
@@ -231,7 +280,7 @@ func createStage(rc *config.Receiver, tmpl *template.Template, wait func() time.
 type RoutingStage map[string]Stage
 
 // Exec implements the Stage interface.
-func (rs RoutingStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+func (rs RoutingStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
 	receiver, ok := ReceiverName(ctx)
 	if !ok {
 		return ctx, nil, fmt.Errorf("receiver missing")
@@ -242,21 +291,21 @@ func (rs RoutingStage) Exec(ctx context.Context, alerts ...*types.Alert) (contex
 		return ctx, nil, fmt.Errorf("stage for receiver missing")
 	}
 
-	return s.Exec(ctx, alerts...)
+	return s.Exec(ctx, l, alerts...)
 }
 
 // A MultiStage executes a series of stages sequencially.
 type MultiStage []Stage
 
 // Exec implements the Stage interface.
-func (ms MultiStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+func (ms MultiStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
 	var err error
 	for _, s := range ms {
 		if len(alerts) == 0 {
 			return ctx, nil, nil
 		}
 
-		ctx, alerts, err = s.Exec(ctx, alerts...)
+		ctx, alerts, err = s.Exec(ctx, l, alerts...)
 		if err != nil {
 			return ctx, nil, err
 		}
@@ -269,7 +318,7 @@ type FanoutStage []Stage
 
 // Exec attempts to execute all stages concurrently and discards the results.
 // It returns its input alerts and a types.MultiError if one or more stages fail.
-func (fs FanoutStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+func (fs FanoutStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
 	var (
 		wg sync.WaitGroup
 		me types.MultiError
@@ -278,9 +327,9 @@ func (fs FanoutStage) Exec(ctx context.Context, alerts ...*types.Alert) (context
 
 	for _, s := range fs {
 		go func(s Stage) {
-			if _, _, err := s.Exec(ctx, alerts...); err != nil {
+			if _, _, err := s.Exec(ctx, l, alerts...); err != nil {
 				me.Add(err)
-				log.Errorf("Error on notify: %s", err)
+				level.Error(l).Log("msg", "Error on notify", "err", err)
 			}
 			wg.Done()
 		}(s)
@@ -293,32 +342,42 @@ func (fs FanoutStage) Exec(ctx context.Context, alerts ...*types.Alert) (context
 	return ctx, alerts, nil
 }
 
+// GossipSettleStage waits until the Gossip has settled to forward alerts.
+type GossipSettleStage struct {
+	peer *cluster.Peer
+}
+
+// NewGossipSettleStage returns a new GossipSettleStage.
+func NewGossipSettleStage(p *cluster.Peer) *GossipSettleStage {
+	return &GossipSettleStage{peer: p}
+}
+
+func (n *GossipSettleStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+	if n.peer != nil {
+		n.peer.WaitReady()
+	}
+	return ctx, alerts, nil
+}
+
 // InhibitStage filters alerts through an inhibition muter.
 type InhibitStage struct {
-	muter  types.Muter
-	marker types.Marker
+	muter types.Muter
 }
 
 // NewInhibitStage return a new InhibitStage.
-func NewInhibitStage(m types.Muter, mk types.Marker) *InhibitStage {
-	return &InhibitStage{
-		muter:  m,
-		marker: mk,
-	}
+func NewInhibitStage(m types.Muter) *InhibitStage {
+	return &InhibitStage{muter: m}
 }
 
 // Exec implements the Stage interface.
-func (n *InhibitStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+func (n *InhibitStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
 	var filtered []*types.Alert
 	for _, a := range alerts {
-		_, ok := n.marker.Inhibited(a.Fingerprint())
 		// TODO(fabxc): increment total alerts counter.
 		// Do not send the alert if the silencer mutes it.
 		if !n.muter.Mutes(a.Labels) {
 			// TODO(fabxc): increment muted alerts counter.
 			filtered = append(filtered, a)
-			// Store whether a previously inhibited alert is firing again.
-			a.WasInhibited = ok
 		}
 	}
 
@@ -340,26 +399,23 @@ func NewSilenceStage(s *silence.Silences, mk types.Marker) *SilenceStage {
 }
 
 // Exec implements the Stage interface.
-func (n *SilenceStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+func (n *SilenceStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
 	var filtered []*types.Alert
 	for _, a := range alerts {
-		_, ok := n.marker.Silenced(a.Fingerprint())
 		// TODO(fabxc): increment total alerts counter.
 		// Do not send the alert if the silencer mutes it.
 		sils, err := n.silences.Query(
-			silence.QState(silence.StateActive),
+			silence.QState(types.SilenceStateActive),
 			silence.QMatches(a.Labels),
 		)
 		if err != nil {
-			log.Errorf("Querying silences failed: %s", err)
+			level.Error(l).Log("msg", "Querying silences failed", "err", err)
 		}
 
 		if len(sils) == 0 {
 			// TODO(fabxc): increment muted alerts counter.
 			filtered = append(filtered, a)
 			n.marker.SetSilenced(a.Labels.Fingerprint())
-			// Store whether a previously silenced alert is firing again.
-			a.WasSilenced = ok
 		} else {
 			ids := make([]string, len(sils))
 			for i, s := range sils {
@@ -386,7 +442,7 @@ func NewWaitStage(wait func() time.Duration) *WaitStage {
 }
 
 // Exec implements the Stage interface.
-func (ws *WaitStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+func (ws *WaitStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
 	select {
 	case <-time.After(ws.wait()):
 	case <-ctx.Done():
@@ -398,22 +454,22 @@ func (ws *WaitStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.
 // DedupStage filters alerts.
 // Filtering happens based on a notification log.
 type DedupStage struct {
-	nflog        nflog.Log
-	recv         *nflogpb.Receiver
-	sendResolved bool
+	nflog NotificationLog
+	recv  *nflogpb.Receiver
+	conf  notifierConfig
 
 	now  func() time.Time
 	hash func(*types.Alert) uint64
 }
 
 // NewDedupStage wraps a DedupStage that runs against the given notification log.
-func NewDedupStage(l nflog.Log, recv *nflogpb.Receiver, sendResolved bool) *DedupStage {
+func NewDedupStage(i Integration, l NotificationLog, recv *nflogpb.Receiver) *DedupStage {
 	return &DedupStage{
-		nflog:        l,
-		recv:         recv,
-		now:          utcNow,
-		sendResolved: sendResolved,
-		hash:         hashAlert,
+		nflog: l,
+		recv:  recv,
+		conf:  i.conf,
+		now:   utcNow,
+		hash:  hashAlert,
 	}
 }
 
@@ -444,7 +500,7 @@ func hashAlert(a *types.Alert) uint64 {
 
 	names := make(model.LabelNames, 0, len(a.Labels))
 
-	for ln, _ := range a.Labels {
+	for ln := range a.Labels {
 		names = append(names, ln)
 	}
 	sort.Sort(names)
@@ -461,36 +517,38 @@ func hashAlert(a *types.Alert) uint64 {
 	return hash
 }
 
-func allAlertsResolved(alerts []*types.Alert) bool {
-	for _, a := range alerts {
-		if !a.Resolved() {
-			return false
-		}
-	}
-	return true
-}
-
-func (n *DedupStage) needsUpdate(entry *nflogpb.Entry, firing, resolved map[uint64]struct{}, repeat time.Duration) (bool, error) {
+func (n *DedupStage) needsUpdate(entry *nflogpb.Entry, firing, resolved map[uint64]struct{}, repeat time.Duration) bool {
 	// If we haven't notified about the alert group before, notify right away
 	// unless we only have resolved alerts.
 	if entry == nil {
-		return ((len(firing) > 0) || (n.sendResolved && len(resolved) > 0)), nil
+		return len(firing) > 0
 	}
 
 	if !entry.IsFiringSubset(firing) {
-		return true, nil
+		return true
 	}
 
-	if n.sendResolved && !entry.IsResolvedSubset(resolved) {
-		return true, nil
+	// Notify about all alerts being resolved.
+	// This is done irrespective of the send_resolved flag to make sure that
+	// the firing alerts are cleared from the notification log.
+	if len(firing) == 0 {
+		// If the current alert group and last notification contain no firing
+		// alert, it means that some alerts have been fired and resolved during the
+		// last interval. In this case, there is no need to notify the receiver
+		// since it doesn't know about them.
+		return len(entry.FiringAlerts) > 0
+	}
+
+	if n.conf.SendResolved() && !entry.IsResolvedSubset(resolved) {
+		return true
 	}
 
 	// Nothing changed, only notify if the repeat interval has passed.
-	return entry.Timestamp.Before(n.now().Add(-repeat)), nil
+	return entry.Timestamp.Before(n.now().Add(-repeat))
 }
 
 // Exec implements the Stage interface.
-func (n *DedupStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+func (n *DedupStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
 	gkey, ok := GroupKey(ctx)
 	if !ok {
 		return ctx, nil, fmt.Errorf("group key missing")
@@ -532,11 +590,9 @@ func (n *DedupStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.
 	case 1:
 		entry = entries[0]
 	case 2:
-		return ctx, nil, fmt.Errorf("Unexpected entry result size %d", len(entries))
+		return ctx, nil, fmt.Errorf("unexpected entry result size %d", len(entries))
 	}
-	if ok, err := n.needsUpdate(entry, firingSet, resolvedSet, repeatInterval); err != nil {
-		return ctx, nil, err
-	} else if ok {
+	if n.needsUpdate(entry, firingSet, resolvedSet, repeatInterval) {
 		return ctx, alerts, nil
 	}
 	return ctx, nil, nil
@@ -546,17 +602,41 @@ func (n *DedupStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.
 // succeeds. It aborts if the context is canceled or timed out.
 type RetryStage struct {
 	integration Integration
+	groupName   string
 }
 
 // NewRetryStage returns a new instance of a RetryStage.
-func NewRetryStage(i Integration) *RetryStage {
+func NewRetryStage(i Integration, groupName string) *RetryStage {
 	return &RetryStage{
 		integration: i,
+		groupName:   groupName,
 	}
 }
 
 // Exec implements the Stage interface.
-func (r RetryStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+func (r RetryStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+	var sent []*types.Alert
+
+	// If we shouldn't send notifications for resolved alerts, but there are only
+	// resolved alerts, report them all as successfully notified (we still want the
+	// notification log to log them for the next run of DedupStage).
+	if !r.integration.conf.SendResolved() {
+		firing, ok := FiringAlerts(ctx)
+		if !ok {
+			return ctx, nil, fmt.Errorf("firing alerts missing")
+		}
+		if len(firing) == 0 {
+			return ctx, alerts, nil
+		}
+		for _, a := range alerts {
+			if a.Status() != model.AlertResolved {
+				sent = append(sent, a)
+			}
+		}
+	} else {
+		sent = alerts
+	}
+
 	var (
 		i    = 0
 		b    = backoff.NewExponentialBackOff()
@@ -580,11 +660,14 @@ func (r RetryStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.C
 
 		select {
 		case <-tick.C:
-			if retry, err := r.integration.Notify(ctx, alerts...); err != nil {
+			now := time.Now()
+			retry, err := r.integration.Notify(ctx, sent...)
+			notificationLatencySeconds.WithLabelValues(r.integration.name).Observe(time.Since(now).Seconds())
+			if err != nil {
 				numFailedNotifications.WithLabelValues(r.integration.name).Inc()
-				log.Debugf("Notify attempt %d for %q failed: %s", i, r.integration.name, err)
+				level.Debug(l).Log("msg", "Notify attempt failed", "attempt", i, "integration", r.integration.name, "receiver", r.groupName, "err", err)
 				if !retry {
-					return ctx, alerts, fmt.Errorf("Cancelling notify retry for %q due to unrecoverable error: %s", r.integration.name, err)
+					return ctx, alerts, fmt.Errorf("cancelling notify retry for %q due to unrecoverable error: %s", r.integration.name, err)
 				}
 
 				// Save this error to be able to return the last seen error by an
@@ -607,12 +690,12 @@ func (r RetryStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.C
 // SetNotifiesStage sets the notification information about passed alerts. The
 // passed alerts should have already been sent to the receivers.
 type SetNotifiesStage struct {
-	nflog nflog.Log
+	nflog NotificationLog
 	recv  *nflogpb.Receiver
 }
 
 // NewSetNotifiesStage returns a new instance of a SetNotifiesStage.
-func NewSetNotifiesStage(l nflog.Log, recv *nflogpb.Receiver) *SetNotifiesStage {
+func NewSetNotifiesStage(l NotificationLog, recv *nflogpb.Receiver) *SetNotifiesStage {
 	return &SetNotifiesStage{
 		nflog: l,
 		recv:  recv,
@@ -620,7 +703,7 @@ func NewSetNotifiesStage(l nflog.Log, recv *nflogpb.Receiver) *SetNotifiesStage
 }
 
 // Exec implements the Stage interface.
-func (n SetNotifiesStage) Exec(ctx context.Context, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+func (n SetNotifiesStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
 	gkey, ok := GroupKey(ctx)
 	if !ok {
 		return ctx, nil, fmt.Errorf("group key missing")
