@@ -1,3 +1,16 @@
+// Copyright 2018 Prometheus Team
+// Licensed under the Apache License, Version 2.0 (the "License");
+// you may not use this file except in compliance with the License.
+// You may obtain a copy of the License at
+//
+// http://www.apache.org/licenses/LICENSE-2.0
+//
+// Unless required by applicable law or agreed to in writing, software
+// distributed under the License is distributed on an "AS IS" BASIS,
+// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+// See the License for the specific language governing permissions and
+// limitations under the License.
+
 package parse
 
 import (
@@ -9,7 +22,7 @@ import (
 )
 
 var (
-	re      = regexp.MustCompile(`(?:\s?)(\w+)(=|=~|!=|!~)(?:\"([^"=~!]+)\"|([^"=~!]+))`)
+	re      = regexp.MustCompile(`(?:\s?)(\w+)(=|=~|!=|!~)(?:\"([^"=~!]+)\"|([^"=~!]+)|\"\")`)
 	typeMap = map[string]labels.MatchType{
 		"=":  labels.MatchEqual,
 		"!=": labels.MatchNotEqual,
@@ -27,13 +40,31 @@ func Matchers(s string) ([]*labels.Matcher, error) {
 		s = s[:len(s)-1]
 	}
 
-	for _, toParse := range strings.Split(s, ",") {
-		m, err := Matcher(toParse)
+	var insideQuotes bool
+	var token string
+	var tokens []string
+	for _, r := range s {
+		if !insideQuotes && r == ',' {
+			tokens = append(tokens, token)
+			token = ""
+			continue
+		}
+		token += string(r)
+		if r == '"' {
+			insideQuotes = !insideQuotes
+		}
+	}
+	if token != "" {
+		tokens = append(tokens, token)
+	}
+	for _, token := range tokens {
+		m, err := Matcher(token)
 		if err != nil {
 			return nil, err
 		}
 		matchers = append(matchers, m)
 	}
+
 	return matchers, nil
 }
 
@@ -53,7 +84,7 @@ func Matcher(s string) (*labels.Matcher, error) {
 func Input(s string) (name, value string, matchType labels.MatchType, err error) {
 	ms := re.FindStringSubmatch(s)
 	if len(ms) < 4 {
-		return "", "", labels.MatchEqual, fmt.Errorf("bad matcher format")
+		return "", "", labels.MatchEqual, fmt.Errorf("bad matcher format: %s", s)
 	}
 
 	var prs bool
@@ -66,7 +97,7 @@ func Input(s string) (name, value string, matchType labels.MatchType, err error)
 		value = ms[4]
 	}
 
-	if name == "" || value == "" || !prs {
+	if name == "" || !prs {
 		return "", "", labels.MatchEqual, fmt.Errorf("failed to parse")
 	}
 
