@@ -17,28 +17,33 @@ package silence
 
 import (
 	"bytes"
-	"errors"
 	"fmt"
 	"io"
 	"math/rand"
 	"os"
+	"reflect"
 	"regexp"
 	"sync"
 	"time"
 
+	"github.com/go-kit/kit/log"
+	"github.com/go-kit/kit/log/level"
 	"github.com/matttproud/golang_protobuf_extensions/pbutil"
+	"github.com/pkg/errors"
+	"github.com/prometheus/alertmanager/cluster"
 	pb "github.com/prometheus/alertmanager/silence/silencepb"
 	"github.com/prometheus/alertmanager/types"
 	"github.com/prometheus/client_golang/prometheus"
-	"github.com/prometheus/common/log"
 	"github.com/prometheus/common/model"
 	"github.com/satori/go.uuid"
-	"github.com/weaveworks/mesh"
 )
 
 // ErrNotFound is returned if a silence was not found.
 var ErrNotFound = fmt.Errorf("not found")
 
+// ErrInvalidState is returned if the state isn't valid.
+var ErrInvalidState = fmt.Errorf("invalid state")
+
 func utcNow() time.Time {
 	return time.Now().UTC()
 }
@@ -94,27 +99,43 @@ type Silences struct {
 	now       func() time.Time
 	retention time.Duration
 
-	gossip mesh.Gossip // gossip channel for sharing silences
-
-	// We store silences in a map of IDs for now. Currently, the memory
-	// state is equivalent to the mesh.GossipData representation.
-	// In the future we'll want support for efficient queries by time
-	// range and affected labels.
-	// Mutex also guards the matcherCache, which always need write lock access.
-	mtx sync.Mutex
-	st  gossipData
-	mc  matcherCache
+	mtx       sync.RWMutex
+	st        state
+	broadcast func([]byte)
+	mc        matcherCache
 }
 
 type metrics struct {
-	gcDuration       prometheus.Summary
-	snapshotDuration prometheus.Summary
-	queriesTotal     prometheus.Counter
-	queryErrorsTotal prometheus.Counter
-	queryDuration    prometheus.Histogram
+	gcDuration              prometheus.Summary
+	snapshotDuration        prometheus.Summary
+	snapshotSize            prometheus.Gauge
+	queriesTotal            prometheus.Counter
+	queryErrorsTotal        prometheus.Counter
+	queryDuration           prometheus.Histogram
+	silencesActive          prometheus.GaugeFunc
+	silencesPending         prometheus.GaugeFunc
+	silencesExpired         prometheus.GaugeFunc
+	propagatedMessagesTotal prometheus.Counter
 }
 
-func newMetrics(r prometheus.Registerer) *metrics {
+func newSilenceMetricByState(s *Silences, st types.SilenceState) prometheus.GaugeFunc {
+	return prometheus.NewGaugeFunc(
+		prometheus.GaugeOpts{
+			Name:        "alertmanager_silences",
+			Help:        "How many silences by state.",
+			ConstLabels: prometheus.Labels{"state": string(st)},
+		},
+		func() float64 {
+			count, err := s.CountState(st)
+			if err != nil {
+				level.Error(s.logger).Log("msg", "Counting silences failed", "err", err)
+			}
+			return float64(count)
+		},
+	)
+}
+
+func newMetrics(r prometheus.Registerer, s *Silences) *metrics {
 	m := &metrics{}
 
 	m.gcDuration = prometheus.NewSummary(prometheus.SummaryOpts{
@@ -125,6 +146,10 @@ func newMetrics(r prometheus.Registerer) *metrics {
 		Name: "alertmanager_silences_snapshot_duration_seconds",
 		Help: "Duration of the last silence snapshot.",
 	})
+	m.snapshotSize = prometheus.NewGauge(prometheus.GaugeOpts{
+		Name: "alertmanager_silences_snapshot_size_bytes",
+		Help: "Size of the last silence snapshot in bytes.",
+	})
 	m.queriesTotal = prometheus.NewCounter(prometheus.CounterOpts{
 		Name: "alertmanager_silences_queries_total",
 		Help: "How many silence queries were received.",
@@ -137,14 +162,28 @@ func newMetrics(r prometheus.Registerer) *metrics {
 		Name: "alertmanager_silences_query_duration_seconds",
 		Help: "Duration of silence query evaluation.",
 	})
+	m.propagatedMessagesTotal = prometheus.NewCounter(prometheus.CounterOpts{
+		Name: "alertmanager_silences_gossip_messages_propagated_total",
+		Help: "Number of received gossip messages that have been further gossiped.",
+	})
+	if s != nil {
+		m.silencesActive = newSilenceMetricByState(s, types.SilenceStateActive)
+		m.silencesPending = newSilenceMetricByState(s, types.SilenceStatePending)
+		m.silencesExpired = newSilenceMetricByState(s, types.SilenceStateExpired)
+	}
 
 	if r != nil {
 		r.MustRegister(
 			m.gcDuration,
 			m.snapshotDuration,
+			m.snapshotSize,
 			m.queriesTotal,
 			m.queryErrorsTotal,
 			m.queryDuration,
+			m.silencesActive,
+			m.silencesPending,
+			m.silencesExpired,
+			m.propagatedMessagesTotal,
 		)
 	}
 	return m
@@ -162,9 +201,6 @@ type Options struct {
 	// garbage collected after the given duration after they ended.
 	Retention time.Duration
 
-	// A function creating a mesh.Gossip on being called with a mesh.Gossiper.
-	Gossip func(g mesh.Gossiper) mesh.Gossip
-
 	// A logger used by background processing.
 	Logger  log.Logger
 	Metrics prometheus.Registerer
@@ -194,18 +230,16 @@ func New(o Options) (*Silences, error) {
 	s := &Silences{
 		mc:        matcherCache{},
 		logger:    log.NewNopLogger(),
-		metrics:   newMetrics(o.Metrics),
 		retention: o.Retention,
 		now:       utcNow,
-		gossip:    nopGossip{},
-		st:        gossipData{},
+		broadcast: func([]byte) {},
+		st:        state{},
 	}
+	s.metrics = newMetrics(o.Metrics, s)
+
 	if o.Logger != nil {
 		s.logger = o.Logger
 	}
-	if o.Gossip != nil {
-		s.gossip = o.Gossip(gossiper{s})
-	}
 	if o.SnapshotReader != nil {
 		if err := s.loadSnapshot(o.SnapshotReader); err != nil {
 			return s, err
@@ -214,11 +248,6 @@ func New(o Options) (*Silences, error) {
 	return s, nil
 }
 
-type nopGossip struct{}
-
-func (nopGossip) GossipBroadcast(d mesh.GossipData)         {}
-func (nopGossip) GossipUnicast(mesh.PeerName, []byte) error { return nil }
-
 // Maintenance garbage collects the silence state at the given interval. If the snapshot
 // file is set, a snapshot is written to it afterwards.
 // Terminates on receiving from stopc.
@@ -228,8 +257,13 @@ func (s *Silences) Maintenance(interval time.Duration, snapf string, stopc <-cha
 
 	f := func() error {
 		start := s.now()
-		s.logger.Info("running maintenance")
-		defer s.logger.With("duration", s.now().Sub(start)).Info("maintenance done")
+		var size int64
+
+		level.Debug(s.logger).Log("msg", "Running maintenance")
+		defer func() {
+			level.Debug(s.logger).Log("msg", "Maintenance done", "duration", s.now().Sub(start), "size", size)
+			s.metrics.snapshotSize.Set(float64(size))
+		}()
 
 		if _, err := s.GC(); err != nil {
 			return err
@@ -241,8 +275,7 @@ func (s *Silences) Maintenance(interval time.Duration, snapf string, stopc <-cha
 		if err != nil {
 			return err
 		}
-		// TODO(fabxc): potentially expose snapshot size in log message.
-		if _, err := s.Snapshot(f); err != nil {
+		if size, err = s.Snapshot(f); err != nil {
 			return err
 		}
 		return f.Close()
@@ -255,7 +288,7 @@ Loop:
 			break Loop
 		case <-t.C:
 			if err := f(); err != nil {
-				s.logger.With("err", err).Error("running maintenance failed")
+				level.Info(s.logger).Log("msg", "Running maintenance failed", "err", err)
 			}
 		}
 	}
@@ -264,7 +297,7 @@ Loop:
 		return
 	}
 	if err := f(); err != nil {
-		s.logger.With("err", err).Info("msg", "creating shutdown snapshot failed")
+		level.Info(s.logger).Log("msg", "Creating shutdown snapshot failed", "err", err)
 	}
 }
 
@@ -290,6 +323,7 @@ func (s *Silences) GC() (int, error) {
 			n++
 		}
 	}
+
 	return n, nil
 }
 
@@ -354,127 +388,115 @@ func (s *Silences) getSilence(id string) (*pb.Silence, bool) {
 }
 
 func (s *Silences) setSilence(sil *pb.Silence) error {
+	sil.UpdatedAt = s.now()
+
+	if err := validateSilence(sil); err != nil {
+		return errors.Wrap(err, "silence invalid")
+	}
+
 	msil := &pb.MeshSilence{
 		Silence:   sil,
 		ExpiresAt: sil.EndsAt.Add(s.retention),
 	}
-	st := gossipData{sil.Id: msil}
+	b, err := marshalMeshSilence(msil)
+	if err != nil {
+		return err
+	}
 
-	s.st.Merge(st)
-	s.gossip.GossipBroadcast(st)
+	s.st.merge(msil)
+	s.broadcast(b)
 
 	return nil
 }
 
-// Create adds a new silence and returns its ID.
-func (s *Silences) Create(sil *pb.Silence) (id string, err error) {
-	if sil.Id != "" {
-		return "", fmt.Errorf("unexpected ID in new silence")
-	}
-	sil.Id = uuid.NewV4().String()
-
-	now := s.now()
-
-	if sil.StartsAt.IsZero() {
-		sil.StartsAt = now
-	} else if sil.StartsAt.Before(now) {
-		return "", fmt.Errorf("new silence must not start in the past")
-	}
-	sil.UpdatedAt = now
-
-	if err := validateSilence(sil); err != nil {
-		return "", fmt.Errorf("invalid silence: %s", err)
-	}
-
+// Set the specified silence. If a silence with the ID already exists and the modification
+// modifies history, the old silence gets expired and a new one is created.
+func (s *Silences) Set(sil *pb.Silence) (string, error) {
 	s.mtx.Lock()
 	defer s.mtx.Unlock()
 
-	if err := s.setSilence(sil); err != nil {
-		return "", err
+	now := s.now()
+	prev, ok := s.getSilence(sil.Id)
+
+	if sil.Id != "" && !ok {
+		return "", ErrNotFound
 	}
-	return sil.Id, nil
+	if ok {
+		if canUpdate(prev, sil, now) {
+			return sil.Id, s.setSilence(sil)
+		}
+		if getState(prev, s.now()) != types.SilenceStateExpired {
+			// We cannot update the silence, expire the old one.
+			if err := s.expire(prev.Id); err != nil {
+				return "", errors.Wrap(err, "expire previous silence")
+			}
+		}
+	}
+	// If we got here it's either a new silence or a replacing one.
+	sil.Id = uuid.NewV4().String()
+
+	if sil.StartsAt.Before(now) {
+		sil.StartsAt = now
+	}
+
+	return sil.Id, s.setSilence(sil)
+}
+
+// canUpdate returns true if silence a can be updated to b without
+// affecting the historic view of silencing.
+func canUpdate(a, b *pb.Silence, now time.Time) bool {
+	if !reflect.DeepEqual(a.Matchers, b.Matchers) {
+		return false
+	}
+	// Allowed timestamp modifications depend on the current time.
+	switch st := getState(a, now); st {
+	case types.SilenceStateActive:
+		if !b.StartsAt.Equal(a.StartsAt) {
+			return false
+		}
+		if b.EndsAt.Before(now) {
+			return false
+		}
+	case types.SilenceStatePending:
+		if b.StartsAt.Before(now) {
+			return false
+		}
+	case types.SilenceStateExpired:
+		return false
+	default:
+		panic("unknown silence state")
+	}
+	return true
 }
 
 // Expire the silence with the given ID immediately.
 func (s *Silences) Expire(id string) error {
 	s.mtx.Lock()
 	defer s.mtx.Unlock()
+	return s.expire(id)
+}
 
+// Expire the silence with the given ID immediately.
+func (s *Silences) expire(id string) error {
 	sil, ok := s.getSilence(id)
 	if !ok {
 		return ErrNotFound
 	}
-
-	now := s.now()
-
-	sil, err := silenceSetTimeRange(sil, now, sil.StartsAt, now)
-	if err != nil {
-		return err
-	}
-	return s.setSilence(sil)
-}
-
-// SetTimeRange adjust the time range of a silence if allowed. If start or end
-// are zero times, the current value remains unmodified.
-func (s *Silences) SetTimeRange(id string, start, end time.Time) error {
-	now := s.now()
-
-	s.mtx.Lock()
-	defer s.mtx.Unlock()
-
-	sil, ok := s.getSilence(id)
-	if !ok {
-		return ErrNotFound
-	}
-
-	if start.IsZero() {
-		start = sil.StartsAt
-	}
-	if end.IsZero() {
-		end = sil.EndsAt
-	}
-
-	sil, err := silenceSetTimeRange(sil, now, start, end)
-	if err != nil {
-		return err
-	}
-	return s.setSilence(sil)
-}
-
-func silenceSetTimeRange(sil *pb.Silence, now, start, end time.Time) (*pb.Silence, error) {
-	if end.Before(start) {
-		return nil, fmt.Errorf("end time must not be before start time")
-	}
-	// Validate modification based on current silence state.
-	switch st := getState(sil, now); st {
-	case StateActive:
-		if !start.Equal(sil.StartsAt) {
-			return nil, fmt.Errorf("start time of active silence cannot be modified")
-		}
-		if end.Before(now) {
-			return nil, fmt.Errorf("end time cannot be set into the past")
-		}
-	case StatePending:
-		if start.Before(now) {
-			return nil, fmt.Errorf("start time cannot be set into the past")
-		}
-	case StateExpired:
-		return nil, fmt.Errorf("expired silence must not be modified")
-	default:
-		return nil, fmt.Errorf("unknown silence state %v", st)
-	}
-
 	sil = cloneSilence(sil)
-	sil.StartsAt = start
-	sil.EndsAt = end
-	sil.UpdatedAt = now
+	now := s.now()
 
-	return sil, nil
-}
+	switch getState(sil, now) {
+	case types.SilenceStateExpired:
+		return errors.Errorf("silence %s already expired", id)
+	case types.SilenceStateActive:
+		sil.EndsAt = now
+	case types.SilenceStatePending:
+		// Set both to now to make Silence move to "expired" state
+		sil.StartsAt = now
+		sil.EndsAt = now
+	}
 
-// AddComment adds a new comment to the silence with the given ID.
-func (s *Silences) AddComment(id string, author, comment string) error {
-	panic("not implemented")
+	return s.setSilence(sil)
 }
 
 // QueryParam expresses parameters along which silences are queried.
@@ -523,29 +545,19 @@ func QMatches(set model.LabelSet) QueryParam {
 	}
 }
 
-// SilenceState describes the state of a silence based on its time range.
-type SilenceState string
-
-// The only possible states of a silence w.r.t a timestamp.
-const (
-	StateActive  SilenceState = "active"
-	StatePending              = "pending"
-	StateExpired              = "expired"
-)
-
 // getState returns a silence's SilenceState at the given timestamp.
-func getState(sil *pb.Silence, ts time.Time) SilenceState {
+func getState(sil *pb.Silence, ts time.Time) types.SilenceState {
 	if ts.Before(sil.StartsAt) {
-		return StatePending
+		return types.SilenceStatePending
 	}
 	if ts.After(sil.EndsAt) {
-		return StateExpired
+		return types.SilenceStateExpired
 	}
-	return StateActive
+	return types.SilenceStateActive
 }
 
 // QState filters queried silences by the given states.
-func QState(states ...SilenceState) QueryParam {
+func QState(states ...types.SilenceState) QueryParam {
 	return func(q *query) error {
 		f := func(sil *pb.Silence, _ *Silences, now time.Time) (bool, error) {
 			s := getState(sil, now)
@@ -562,6 +574,19 @@ func QState(states ...SilenceState) QueryParam {
 	}
 }
 
+// QueryOne queries with the given parameters and returns the first result.
+// Returns ErrNotFound if the query result is empty.
+func (s *Silences) QueryOne(params ...QueryParam) (*pb.Silence, error) {
+	res, err := s.Query(params...)
+	if err != nil {
+		return nil, err
+	}
+	if len(res) == 0 {
+		return nil, ErrNotFound
+	}
+	return res[0], nil
+}
+
 // Query for silences based on the given query parameters.
 func (s *Silences) Query(params ...QueryParam) ([]*pb.Silence, error) {
 	start := time.Now()
@@ -583,6 +608,16 @@ func (s *Silences) Query(params ...QueryParam) ([]*pb.Silence, error) {
 	return sils, err
 }
 
+// Count silences by state.
+func (s *Silences) CountState(states ...types.SilenceState) (int, error) {
+	// This could probably be optimized.
+	sils, err := s.Query(QState(states...))
+	if err != nil {
+		return -1, err
+	}
+	return len(sils), nil
+}
+
 func (s *Silences) query(q *query, now time.Time) ([]*pb.Silence, error) {
 	// If we have an ID constraint, all silences are our base set.
 	// This and the use of post-filter functions is the
@@ -594,7 +629,7 @@ func (s *Silences) query(q *query, now time.Time) ([]*pb.Silence, error) {
 
 	if q.ids != nil {
 		for _, id := range q.ids {
-			if s, ok := s.st[string(id)]; ok {
+			if s, ok := s.st[id]; ok {
 				res = append(res, s.Silence)
 			}
 		}
@@ -618,7 +653,7 @@ func (s *Silences) query(q *query, now time.Time) ([]*pb.Silence, error) {
 			}
 		}
 		if !remove {
-			resf = append(resf, sil)
+			resf = append(resf, cloneSilence(sil))
 		}
 	}
 
@@ -628,183 +663,137 @@ func (s *Silences) query(q *query, now time.Time) ([]*pb.Silence, error) {
 // loadSnapshot loads a snapshot generated by Snapshot() into the state.
 // Any previous state is wiped.
 func (s *Silences) loadSnapshot(r io.Reader) error {
-	st := gossipData{}
-
-	s.mtx.Lock()
-	defer s.mtx.Unlock()
-
-	for {
-		var sil pb.MeshSilence
-		if _, err := pbutil.ReadDelimited(r, &sil); err != nil {
-			if err == io.EOF {
-				break
-			}
-			return err
-		}
-		st[sil.Silence.Id] = &sil
-		_, err := s.mc.Get(sil.Silence)
-		if err != nil {
-			return err
-		}
+	st, err := decodeState(r)
+	if err != nil {
+		return err
 	}
-
+	for _, e := range st {
+		// Comments list was moved to a single comment. Upgrade on loading the snapshot.
+		if len(e.Silence.Comments) > 0 {
+			e.Silence.Comment = e.Silence.Comments[0].Comment
+			e.Silence.CreatedBy = e.Silence.Comments[0].Author
+			e.Silence.Comments = nil
+		}
+		st[e.Silence.Id] = e
+	}
+	s.mtx.Lock()
 	s.st = st
+	s.mtx.Unlock()
 
 	return nil
 }
 
 // Snapshot writes the full internal state into the writer and returns the number of bytes
 // written.
-func (s *Silences) Snapshot(w io.Writer) (int, error) {
+func (s *Silences) Snapshot(w io.Writer) (int64, error) {
 	start := time.Now()
 	defer func() { s.metrics.snapshotDuration.Observe(time.Since(start).Seconds()) }()
 
+	s.mtx.RLock()
+	defer s.mtx.RUnlock()
+
+	b, err := s.st.MarshalBinary()
+	if err != nil {
+		return 0, err
+	}
+
+	return io.Copy(w, bytes.NewReader(b))
+}
+
+// MarshalBinary serializes all silences.
+func (s *Silences) MarshalBinary() ([]byte, error) {
 	s.mtx.Lock()
 	defer s.mtx.Unlock()
 
-	var n int
-	for _, s := range s.st {
-		m, err := pbutil.WriteDelimited(w, s)
-		if err != nil {
-			return n + m, err
+	return s.st.MarshalBinary()
+}
+
+// Merge merges silence state received from the cluster with the local state.
+func (s *Silences) Merge(b []byte) error {
+	st, err := decodeState(bytes.NewReader(b))
+	if err != nil {
+		return err
+	}
+	s.mtx.Lock()
+	defer s.mtx.Unlock()
+
+	for _, e := range st {
+		if merged := s.st.merge(e); merged && !cluster.OversizedMessage(b) {
+			// If this is the first we've seen the message and it's
+			// not oversized, gossip it to other nodes. We don't
+			// propagate oversized messages because they're sent to
+			// all nodes already.
+			s.broadcast(b)
+			s.metrics.propagatedMessagesTotal.Inc()
+			level.Debug(s.logger).Log("msg", "gossiping new silence", "silence", e)
 		}
-		n += m
 	}
-	return n, nil
+	return nil
 }
 
-type gossiper struct {
-	*Silences
+func (s *Silences) SetBroadcast(f func([]byte)) {
+	s.mtx.Lock()
+	s.broadcast = f
+	s.mtx.Unlock()
 }
 
-// Gossip implements the mesh.Gossiper interface.
-func (g gossiper) Gossip() mesh.GossipData {
-	g.mtx.Lock()
-	defer g.mtx.Unlock()
+type state map[string]*pb.MeshSilence
 
-	return g.st.clone()
-}
-
-// OnGossip implements the mesh.Gossiper interface.
-func (g gossiper) OnGossip(msg []byte) (mesh.GossipData, error) {
-	gd, err := decodeGossipData(msg)
-	if err != nil {
-		return nil, err
+func (s state) merge(e *pb.MeshSilence) bool {
+	// Comments list was moved to a single comment. Apply upgrade
+	// on silences received from peers.
+	if len(e.Silence.Comments) > 0 {
+		e.Silence.Comment = e.Silence.Comments[0].Comment
+		e.Silence.CreatedBy = e.Silence.Comments[0].Author
+		e.Silence.Comments = nil
 	}
-	g.mtx.Lock()
-	defer g.mtx.Unlock()
+	id := e.Silence.Id
 
-	if delta := g.st.mergeDelta(gd); len(delta) > 0 {
-		return delta, nil
+	prev, ok := s[id]
+	if !ok || prev.Silence.UpdatedAt.Before(e.Silence.UpdatedAt) {
+		s[id] = e
+		return true
 	}
-	return nil, nil
+	return false
 }
 
-// OnGossipBroadcast implements the mesh.Gossiper interface.
-func (g gossiper) OnGossipBroadcast(src mesh.PeerName, msg []byte) (mesh.GossipData, error) {
-	gd, err := decodeGossipData(msg)
-	if err != nil {
-		return nil, err
+func (s state) MarshalBinary() ([]byte, error) {
+	var buf bytes.Buffer
+
+	for _, e := range s {
+		if _, err := pbutil.WriteDelimited(&buf, e); err != nil {
+			return nil, err
+		}
 	}
-	g.mtx.Lock()
-	defer g.mtx.Unlock()
-
-	return g.st.mergeDelta(gd), nil
+	return buf.Bytes(), nil
 }
 
-// OnGossipUnicast implements the mesh.Gossiper interface.
-// It always panics.
-func (g gossiper) OnGossipUnicast(src mesh.PeerName, msg []byte) error {
-	panic("not implemented")
-}
-
-type gossipData map[string]*pb.MeshSilence
-
-func decodeGossipData(msg []byte) (gossipData, error) {
-	gd := gossipData{}
-	rd := bytes.NewReader(msg)
-
+func decodeState(r io.Reader) (state, error) {
+	st := state{}
 	for {
 		var s pb.MeshSilence
-		if _, err := pbutil.ReadDelimited(rd, &s); err != nil {
-			if err == io.EOF {
-				break
+		_, err := pbutil.ReadDelimited(r, &s)
+		if err == nil {
+			if s.Silence == nil {
+				return nil, ErrInvalidState
 			}
-			return gd, err
-		}
-		gd[s.Silence.Id] = &s
-	}
-	return gd, nil
-}
-
-// Encode implements the mesh.GossipData interface.
-func (gd gossipData) Encode() [][]byte {
-	// Split into sub-messages of ~1MB.
-	const maxSize = 1024 * 1024
-
-	var (
-		buf bytes.Buffer
-		res [][]byte
-		n   int
-	)
-	for _, s := range gd {
-		m, err := pbutil.WriteDelimited(&buf, s)
-		n += m
-		if err != nil {
-			// TODO(fabxc): log error and skip entry. Or can this really not happen with a bytes.Buffer?
-			panic(err)
-		}
-		if n > maxSize {
-			res = append(res, buf.Bytes())
-			buf = bytes.Buffer{}
-		}
-	}
-	if buf.Len() > 0 {
-		res = append(res, buf.Bytes())
-	}
-	return res
-}
-
-func (gd gossipData) clone() gossipData {
-	res := make(gossipData, len(gd))
-	for id, s := range gd {
-		res[id] = s
-	}
-	return res
-}
-
-// Merge the silence set with gossip data and return a new silence state.
-func (gd gossipData) Merge(other mesh.GossipData) mesh.GossipData {
-	for id, s := range other.(gossipData) {
-		prev, ok := gd[id]
-		if !ok {
-			gd[id] = s
+			st[s.Silence.Id] = &s
 			continue
 		}
-		if prev.Silence.UpdatedAt.Before(s.Silence.UpdatedAt) {
-			gd[id] = s
+		if err == io.EOF {
+			break
 		}
+		return nil, err
 	}
-	return gd
+	return st, nil
 }
 
-// mergeDelta behaves like Merge but returns a gossipData only
-// containing things that have changed.
-func (gd gossipData) mergeDelta(od gossipData) gossipData {
-	delta := gossipData{}
-	for id, s := range od {
-		prev, ok := gd[id]
-		if !ok {
-			gd[id] = s
-			delta[id] = s
-			continue
-		}
-		if prev.Silence.UpdatedAt.Before(s.Silence.UpdatedAt) {
-			gd[id] = s
-			delta[id] = s
-		}
+func marshalMeshSilence(e *pb.MeshSilence) ([]byte, error) {
+	var buf bytes.Buffer
+	if _, err := pbutil.WriteDelimited(&buf, e); err != nil {
+		return nil, err
 	}
-	return delta
+	return buf.Bytes(), nil
 }
 
 // replaceFile wraps a file that is moved to another filename on closing.
