@@ -22,10 +22,11 @@ import (
 	"testing"
 	"time"
 
+	"github.com/matttproud/golang_protobuf_extensions/pbutil"
 	pb "github.com/prometheus/alertmanager/silence/silencepb"
+	"github.com/prometheus/alertmanager/types"
 	"github.com/prometheus/common/model"
 	"github.com/stretchr/testify/require"
-	"github.com/weaveworks/mesh"
 )
 
 func TestOptionsValidate(t *testing.T) {
@@ -80,12 +81,12 @@ func TestSilencesGC(t *testing.T) {
 	newSilence := func(exp time.Time) *pb.MeshSilence {
 		return &pb.MeshSilence{ExpiresAt: exp}
 	}
-	s.st = gossipData{
+	s.st = state{
 		"1": newSilence(now),
 		"2": newSilence(now.Add(-time.Second)),
 		"3": newSilence(now.Add(time.Second)),
 	}
-	want := gossipData{
+	want := state{
 		"3": newSilence(now.Add(time.Second)),
 	}
 
@@ -137,7 +138,7 @@ func TestSilencesSnapshot(t *testing.T) {
 		f, err := ioutil.TempFile("", "snapshot")
 		require.NoError(t, err, "creating temp file failed")
 
-		s1 := &Silences{st: gossipData{}, metrics: newMetrics(nil)}
+		s1 := &Silences{st: state{}, metrics: newMetrics(nil, nil)}
 		// Setup internal state manually.
 		for _, e := range c.entries {
 			s1.st[e.Silence.Id] = e
@@ -151,7 +152,7 @@ func TestSilencesSnapshot(t *testing.T) {
 		require.NoError(t, err, "opening snapshot file failed")
 
 		// Check again against new nlog instance.
-		s2 := &Silences{mc: matcherCache{}}
+		s2 := &Silences{mc: matcherCache{}, st: state{}}
 		err = s2.loadSnapshot(f)
 		require.NoError(t, err, "error loading snapshot")
 		require.Equal(t, s1.st, s2.st, "state after loading snapshot did not match snapshotted state")
@@ -160,13 +161,6 @@ func TestSilencesSnapshot(t *testing.T) {
 	}
 }
 
-type mockGossip struct {
-	broadcast func(mesh.GossipData)
-}
-
-func (g *mockGossip) GossipBroadcast(d mesh.GossipData)         { g.broadcast(d) }
-func (g *mockGossip) GossipUnicast(mesh.PeerName, []byte) error { panic("not implemented") }
-
 func TestSilencesSetSilence(t *testing.T) {
 	s, err := New(Options{
 		Retention: time.Minute,
@@ -177,39 +171,55 @@ func TestSilencesSetSilence(t *testing.T) {
 	nowpb := now
 
 	sil := &pb.Silence{
-		Id:     "some_id",
-		EndsAt: nowpb,
+		Id:       "some_id",
+		Matchers: []*pb.Matcher{{Name: "abc", Pattern: "def"}},
+		StartsAt: nowpb,
+		EndsAt:   nowpb,
 	}
 
-	want := gossipData{
+	want := state{
 		"some_id": &pb.MeshSilence{
 			Silence:   sil,
 			ExpiresAt: now.Add(time.Minute),
 		},
 	}
 
-	var called bool
-	s.gossip = &mockGossip{
-		broadcast: func(d mesh.GossipData) {
-			data, ok := d.(gossipData)
-			require.True(t, ok, "gossip data of unknown type")
-			require.Equal(t, want, data, "unexpected gossip broadcast data")
+	done := make(chan bool)
+	s.broadcast = func(b []byte) {
+		var e pb.MeshSilence
+		r := bytes.NewReader(b)
+		_, err := pbutil.ReadDelimited(r, &e)
+		require.NoError(t, err)
 
-			called = true
-		},
+		require.Equal(t, want["some_id"], &e)
+		close(done)
 	}
-	require.NoError(t, s.setSilence(sil))
-	require.True(t, called, "GossipBroadcast was not called")
+
+	// setSilence() is always called with s.mtx locked()
+	go func() {
+		s.mtx.Lock()
+		require.NoError(t, s.setSilence(sil))
+		s.mtx.Unlock()
+	}()
+
+	// GossipBroadcast is called in a goroutine.
+	select {
+	case <-done:
+	case <-time.After(1 * time.Second):
+		t.Fatal("GossipBroadcast was not called")
+	}
+
 	require.Equal(t, want, s.st, "Unexpected silence state")
 }
 
-func TestSilenceCreate(t *testing.T) {
+func TestSilenceSet(t *testing.T) {
 	s, err := New(Options{
 		Retention: time.Hour,
 	})
 	require.NoError(t, err)
 
 	now := utcNow()
+	now1 := now
 	s.now = func() time.Time { return now }
 
 	// Insert silence with fixed start time.
@@ -218,60 +228,144 @@ func TestSilenceCreate(t *testing.T) {
 		StartsAt: now.Add(2 * time.Minute),
 		EndsAt:   now.Add(5 * time.Minute),
 	}
-	id1, err := s.Create(sil1)
+	id1, err := s.Set(sil1)
 	require.NoError(t, err)
 	require.NotEqual(t, id1, "")
 
-	want := gossipData{
+	want := state{
 		id1: &pb.MeshSilence{
 			Silence: &pb.Silence{
 				Id:        id1,
 				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "b"}},
-				StartsAt:  now.Add(2 * time.Minute),
-				EndsAt:    now.Add(5 * time.Minute),
-				UpdatedAt: now,
+				StartsAt:  now1.Add(2 * time.Minute),
+				EndsAt:    now1.Add(5 * time.Minute),
+				UpdatedAt: now1,
 			},
-			ExpiresAt: now.Add(5*time.Minute + s.retention),
+			ExpiresAt: now1.Add(5*time.Minute + s.retention),
 		},
 	}
 	require.Equal(t, want, s.st, "unexpected state after silence creation")
 
 	// Insert silence with unset start time. Must be set to now.
+	now = now.Add(time.Minute)
+	now2 := now
+
 	sil2 := &pb.Silence{
 		Matchers: []*pb.Matcher{{Name: "a", Pattern: "b"}},
 		EndsAt:   now.Add(1 * time.Minute),
 	}
-	id2, err := s.Create(sil2)
+	id2, err := s.Set(sil2)
 	require.NoError(t, err)
 	require.NotEqual(t, id2, "")
 
-	want = gossipData{
-		id1: &pb.MeshSilence{
-			Silence: &pb.Silence{
-				Id:        id1,
-				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "b"}},
-				StartsAt:  now.Add(2 * time.Minute),
-				EndsAt:    now.Add(5 * time.Minute),
-				UpdatedAt: now,
-			},
-			ExpiresAt: now.Add(5*time.Minute + s.retention),
-		},
+	want = state{
+		id1: want[id1],
 		id2: &pb.MeshSilence{
 			Silence: &pb.Silence{
 				Id:        id2,
 				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "b"}},
-				StartsAt:  now,
-				EndsAt:    now.Add(1 * time.Minute),
-				UpdatedAt: now,
+				StartsAt:  now2,
+				EndsAt:    now2.Add(1 * time.Minute),
+				UpdatedAt: now2,
 			},
-			ExpiresAt: now.Add(1*time.Minute + s.retention),
+			ExpiresAt: now2.Add(1*time.Minute + s.retention),
 		},
 	}
 	require.Equal(t, want, s.st, "unexpected state after silence creation")
 
+	// Overwrite silence 2 with new end time.
+	now = now.Add(time.Minute)
+	now3 := now
+
+	sil3 := cloneSilence(sil2)
+	sil3.EndsAt = now.Add(100 * time.Minute)
+
+	id3, err := s.Set(sil3)
+	require.NoError(t, err)
+	require.Equal(t, id2, id3)
+
+	want = state{
+		id1: want[id1],
+		id2: &pb.MeshSilence{
+			Silence: &pb.Silence{
+				Id:        id2,
+				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "b"}},
+				StartsAt:  now2,
+				EndsAt:    now3.Add(100 * time.Minute),
+				UpdatedAt: now3,
+			},
+			ExpiresAt: now3.Add(100*time.Minute + s.retention),
+		},
+	}
+	require.Equal(t, want, s.st, "unexpected state after silence creation")
+
+	// Update silence 2 with new matcher expires it and creates a new one.
+	now = now.Add(time.Minute)
+	now4 := now
+
+	sil4 := cloneSilence(sil3)
+	sil4.Matchers = []*pb.Matcher{{Name: "a", Pattern: "c"}}
+
+	id4, err := s.Set(sil4)
+	require.NoError(t, err)
+	require.NotEqual(t, id2, id4)
+
+	want = state{
+		id1: want[id1],
+		id2: &pb.MeshSilence{
+			Silence: &pb.Silence{
+				Id:        id2,
+				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "b"}},
+				StartsAt:  now2,
+				EndsAt:    now4,
+				UpdatedAt: now4,
+			},
+			ExpiresAt: now4.Add(s.retention),
+		},
+		id4: &pb.MeshSilence{
+			Silence: &pb.Silence{
+				Id:        id4,
+				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "c"}},
+				StartsAt:  now4,
+				EndsAt:    now3.Add(100 * time.Minute),
+				UpdatedAt: now4,
+			},
+			ExpiresAt: now3.Add(100*time.Minute + s.retention),
+		},
+	}
+	require.Equal(t, want, s.st, "unexpected state after silence creation")
+
+	// Re-create expired silence.
+	now = now.Add(time.Minute)
+	now5 := now
+
+	sil5 := cloneSilence(sil3)
+	sil5.StartsAt = now
+	sil5.EndsAt = now.Add(5 * time.Minute)
+
+	id5, err := s.Set(sil5)
+	require.NoError(t, err)
+	require.NotEqual(t, id2, id4)
+
+	want = state{
+		id1: want[id1],
+		id2: want[id2],
+		id4: want[id4],
+		id5: &pb.MeshSilence{
+			Silence: &pb.Silence{
+				Id:        id5,
+				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "b"}},
+				StartsAt:  now5,
+				EndsAt:    now5.Add(5 * time.Minute),
+				UpdatedAt: now5,
+			},
+			ExpiresAt: now5.Add(5*time.Minute + s.retention),
+		},
+	}
+	require.Equal(t, want, s.st, "unexpected state after silence creation")
 }
 
-func TestSilencesCreateFail(t *testing.T) {
+func TestSilencesSetFail(t *testing.T) {
 	s, err := New(Options{})
 	require.NoError(t, err)
 
@@ -284,17 +378,14 @@ func TestSilencesCreateFail(t *testing.T) {
 	}{
 		{
 			s:   &pb.Silence{Id: "some_id"},
-			err: "unexpected ID in new silence",
-		}, {
-			s:   &pb.Silence{StartsAt: now.Add(-time.Minute)},
-			err: "new silence must not start in the past",
+			err: ErrNotFound.Error(),
 		}, {
 			s:   &pb.Silence{}, // Silence without matcher.
-			err: "invalid silence",
+			err: "silence invalid",
 		},
 	}
 	for _, c := range cases {
-		_, err := s.Create(c.s)
+		_, err := s.Set(c.s)
 		if err == nil {
 			if c.err != "" {
 				t.Errorf("expected error containing %q but got none", c.err)
@@ -316,7 +407,7 @@ func TestQState(t *testing.T) {
 
 	cases := []struct {
 		sil    *pb.Silence
-		states []SilenceState
+		states []types.SilenceState
 		keep   bool
 	}{
 		{
@@ -324,7 +415,7 @@ func TestQState(t *testing.T) {
 				StartsAt: now.Add(time.Minute),
 				EndsAt:   now.Add(time.Hour),
 			},
-			states: []SilenceState{StateActive, StateExpired},
+			states: []types.SilenceState{types.SilenceStateActive, types.SilenceStateExpired},
 			keep:   false,
 		},
 		{
@@ -332,7 +423,7 @@ func TestQState(t *testing.T) {
 				StartsAt: now.Add(time.Minute),
 				EndsAt:   now.Add(time.Hour),
 			},
-			states: []SilenceState{StatePending},
+			states: []types.SilenceState{types.SilenceStatePending},
 			keep:   true,
 		},
 		{
@@ -340,7 +431,7 @@ func TestQState(t *testing.T) {
 				StartsAt: now.Add(time.Minute),
 				EndsAt:   now.Add(time.Hour),
 			},
-			states: []SilenceState{StateExpired, StatePending},
+			states: []types.SilenceState{types.SilenceStateExpired, types.SilenceStatePending},
 			keep:   true,
 		},
 	}
@@ -407,7 +498,7 @@ func TestQMatches(t *testing.T) {
 		},
 	}
 	for _, c := range cases {
-		drop, err := f(c.sil, &Silences{mc: matcherCache{}}, time.Time{})
+		drop, err := f(c.sil, &Silences{mc: matcherCache{}, st: state{}}, time.Time{})
 		require.NoError(t, err)
 		require.Equal(t, c.drop, drop, "unexpected filter result")
 	}
@@ -417,7 +508,7 @@ func TestSilencesQuery(t *testing.T) {
 	s, err := New(Options{})
 	require.NoError(t, err)
 
-	s.st = gossipData{
+	s.st = state{
 		"1": &pb.MeshSilence{Silence: &pb.Silence{Id: "1"}},
 		"2": &pb.MeshSilence{Silence: &pb.Silence{Id: "2"}},
 		"3": &pb.MeshSilence{Silence: &pb.Silence{Id: "3"}},
@@ -497,127 +588,216 @@ func (s silencesByID) Len() int           { return len(s) }
 func (s silencesByID) Swap(i, j int)      { s[i], s[j] = s[j], s[i] }
 func (s silencesByID) Less(i, j int) bool { return s[i].Id < s[j].Id }
 
-func TestSilenceSetTimeRange(t *testing.T) {
+func TestSilenceCanUpdate(t *testing.T) {
 	now := utcNow()
 
 	cases := []struct {
-		sil        *pb.Silence
-		start, end time.Time
-		err        string
+		a, b *pb.Silence
+		ok   bool
 	}{
 		// Bad arguments.
 		{
-			sil:   &pb.Silence{},
-			start: now,
-			end:   now.Add(-time.Minute),
-			err:   "end time must not be before start time",
+			a: &pb.Silence{},
+			b: &pb.Silence{
+				StartsAt: now,
+				EndsAt:   now.Add(-time.Minute),
+			},
+			ok: false,
 		},
 		// Expired silence.
 		{
-			sil: &pb.Silence{
+			a: &pb.Silence{
 				StartsAt: now.Add(-time.Hour),
 				EndsAt:   now.Add(-time.Second),
 			},
-			start: now,
-			end:   now,
-			err:   "expired silence must not be modified",
+			b: &pb.Silence{
+				StartsAt: now,
+				EndsAt:   now,
+			},
+			ok: false,
 		},
 		// Pending silences.
 		{
-			sil: &pb.Silence{
+			a: &pb.Silence{
 				StartsAt:  now.Add(time.Hour),
 				EndsAt:    now.Add(2 * time.Hour),
 				UpdatedAt: now.Add(-time.Hour),
 			},
-			start: now.Add(-time.Minute),
-			end:   now.Add(time.Hour),
-			err:   "start time cannot be set into the past",
+			b: &pb.Silence{
+				StartsAt: now.Add(-time.Minute),
+				EndsAt:   now.Add(time.Hour),
+			},
+			ok: false,
 		},
 		{
-			sil: &pb.Silence{
+			a: &pb.Silence{
 				StartsAt:  now.Add(time.Hour),
 				EndsAt:    now.Add(2 * time.Hour),
 				UpdatedAt: now.Add(-time.Hour),
 			},
-			start: now.Add(time.Minute),
-			end:   now.Add(time.Minute),
+			b: &pb.Silence{
+				StartsAt: now.Add(time.Minute),
+				EndsAt:   now.Add(time.Minute),
+			},
+			ok: true,
 		},
 		{
-			sil: &pb.Silence{
+			a: &pb.Silence{
 				StartsAt:  now.Add(time.Hour),
 				EndsAt:    now.Add(2 * time.Hour),
 				UpdatedAt: now.Add(-time.Hour),
 			},
-			start: now, // set to exactly start now.
-			end:   now.Add(2 * time.Hour),
+			b: &pb.Silence{
+				StartsAt: now, // set to exactly start now.
+				EndsAt:   now.Add(2 * time.Hour),
+			},
+			ok: true,
 		},
 		// Active silences.
 		{
-			sil: &pb.Silence{
+			a: &pb.Silence{
 				StartsAt:  now.Add(-time.Hour),
 				EndsAt:    now.Add(2 * time.Hour),
 				UpdatedAt: now.Add(-time.Hour),
 			},
-			start: now.Add(-time.Minute),
-			end:   now.Add(2 * time.Hour),
-			err:   "start time of active silence cannot be modified",
+			b: &pb.Silence{
+				StartsAt: now.Add(-time.Minute),
+				EndsAt:   now.Add(2 * time.Hour),
+			},
+			ok: false,
 		},
 		{
-			sil: &pb.Silence{
+			a: &pb.Silence{
 				StartsAt:  now.Add(-time.Hour),
 				EndsAt:    now.Add(2 * time.Hour),
 				UpdatedAt: now.Add(-time.Hour),
 			},
-			start: now.Add(-time.Hour),
-			end:   now.Add(-time.Second),
-			err:   "end time cannot be set into the past",
+			b: &pb.Silence{
+				StartsAt: now.Add(-time.Hour),
+				EndsAt:   now.Add(-time.Second),
+			},
+			ok: false,
 		},
 		{
-			sil: &pb.Silence{
+			a: &pb.Silence{
 				StartsAt:  now.Add(-time.Hour),
 				EndsAt:    now.Add(2 * time.Hour),
 				UpdatedAt: now.Add(-time.Hour),
 			},
-			start: now.Add(-time.Hour),
-			end:   now,
+			b: &pb.Silence{
+				StartsAt: now.Add(-time.Hour),
+				EndsAt:   now,
+			},
+			ok: true,
 		},
 		{
-			sil: &pb.Silence{
+			a: &pb.Silence{
 				StartsAt:  now.Add(-time.Hour),
 				EndsAt:    now.Add(2 * time.Hour),
 				UpdatedAt: now.Add(-time.Hour),
 			},
-			start: now.Add(-time.Hour),
-			end:   now.Add(3 * time.Hour),
+			b: &pb.Silence{
+				StartsAt: now.Add(-time.Hour),
+				EndsAt:   now.Add(3 * time.Hour),
+			},
+			ok: true,
 		},
 	}
 	for _, c := range cases {
-		origSilence := cloneSilence(c.sil)
-
-		sil, err := silenceSetTimeRange(c.sil, now, c.start, c.end)
-		if err == nil {
-			if c.err != "" {
-				t.Errorf("expected error containing %q but got none", c.err)
-			}
-			// The original silence must not have been modified.
-			require.Equal(t, origSilence, c.sil, "original silence illegally modified")
-
-			require.Equal(t, sil.StartsAt, c.start)
-			require.Equal(t, sil.EndsAt, c.end)
-			require.Equal(t, sil.UpdatedAt, now)
-			continue
+		ok := canUpdate(c.a, c.b, now)
+		if ok && !c.ok {
+			t.Errorf("expected not-updateable but was: %v, %v", c.a, c.b)
 		}
-		if err != nil && c.err == "" {
-			t.Errorf("unexpected error %q", err)
-			continue
+		if ok && !c.ok {
+			t.Errorf("expected updateable but was not: %v, %v", c.a, c.b)
 		}
-		if !strings.Contains(err.Error(), c.err) {
-			t.Errorf("expected error to contain %q but got %q", c.err, err)
-		}
-
 	}
 }
 
+func TestSilenceExpire(t *testing.T) {
+	s, err := New(Options{})
+	require.NoError(t, err)
+
+	now := time.Now()
+	s.now = func() time.Time { return now }
+
+	m := &pb.Matcher{Type: pb.Matcher_EQUAL, Name: "a", Pattern: "b"}
+
+	s.st = state{
+		"pending": &pb.MeshSilence{Silence: &pb.Silence{
+			Id:        "pending",
+			Matchers:  []*pb.Matcher{m},
+			StartsAt:  now.Add(time.Minute),
+			EndsAt:    now.Add(time.Hour),
+			UpdatedAt: now.Add(-time.Hour),
+		}},
+		"active": &pb.MeshSilence{Silence: &pb.Silence{
+			Id:        "active",
+			Matchers:  []*pb.Matcher{m},
+			StartsAt:  now.Add(-time.Minute),
+			EndsAt:    now.Add(time.Hour),
+			UpdatedAt: now.Add(-time.Hour),
+		}},
+		"expired": &pb.MeshSilence{Silence: &pb.Silence{
+			Id:        "expired",
+			Matchers:  []*pb.Matcher{m},
+			StartsAt:  now.Add(-time.Hour),
+			EndsAt:    now.Add(-time.Minute),
+			UpdatedAt: now.Add(-time.Hour),
+		}},
+	}
+
+	count, err := s.CountState(types.SilenceStatePending)
+	require.NoError(t, err)
+	require.Equal(t, 1, count)
+
+	require.NoError(t, s.expire("pending"))
+	require.NoError(t, s.expire("active"))
+
+	err = s.expire("expired")
+	require.Error(t, err)
+	require.Contains(t, err.Error(), "already expired")
+
+	sil, err := s.QueryOne(QIDs("pending"))
+	require.NoError(t, err)
+	require.Equal(t, &pb.Silence{
+		Id:        "pending",
+		Matchers:  []*pb.Matcher{m},
+		StartsAt:  now,
+		EndsAt:    now,
+		UpdatedAt: now,
+	}, sil)
+
+	count, err = s.CountState(types.SilenceStatePending)
+	require.NoError(t, err)
+	require.Equal(t, 0, count)
+
+	// Expiring a pending Silence should make the API return the
+	// SilenceStateExpired Silence state.
+	silenceState := types.CalcSilenceState(sil.StartsAt, sil.EndsAt)
+	require.Equal(t, silenceState, types.SilenceStateExpired)
+
+	sil, err = s.QueryOne(QIDs("active"))
+	require.NoError(t, err)
+	require.Equal(t, &pb.Silence{
+		Id:        "active",
+		Matchers:  []*pb.Matcher{m},
+		StartsAt:  now.Add(-time.Minute),
+		EndsAt:    now,
+		UpdatedAt: now,
+	}, sil)
+
+	sil, err = s.QueryOne(QIDs("expired"))
+	require.NoError(t, err)
+	require.Equal(t, &pb.Silence{
+		Id:        "expired",
+		Matchers:  []*pb.Matcher{m},
+		StartsAt:  now.Add(-time.Hour),
+		EndsAt:    now.Add(-time.Minute),
+		UpdatedAt: now.Add(-time.Hour),
+	}, sil)
+}
+
 func TestValidateMatcher(t *testing.T) {
 	cases := []struct {
 		m   *pb.Matcher
@@ -803,64 +983,51 @@ func TestValidateSilence(t *testing.T) {
 	}
 }
 
-func TestGossipDataMerge(t *testing.T) {
+func TestStateMerge(t *testing.T) {
 	now := utcNow()
 
 	// We only care about key names and timestamps for the
 	// merging logic.
-	newSilence := func(ts time.Time) *pb.MeshSilence {
+	newSilence := func(id string, ts time.Time) *pb.MeshSilence {
 		return &pb.MeshSilence{
-			Silence: &pb.Silence{UpdatedAt: ts},
+			Silence: &pb.Silence{Id: id, UpdatedAt: ts},
 		}
 	}
+
 	cases := []struct {
-		a, b         gossipData
-		final, delta gossipData
+		a, b  state
+		final state
 	}{
 		{
-			a: gossipData{
-				"a1": newSilence(now),
-				"a2": newSilence(now),
-				"a3": newSilence(now),
+			a: state{
+				"a1": newSilence("a1", now),
+				"a2": newSilence("a2", now),
+				"a3": newSilence("a3", now),
 			},
-			b: gossipData{
-				"b1": newSilence(now),                   // new key, should be added
-				"a2": newSilence(now.Add(-time.Minute)), // older timestamp, should be dropped
-				"a3": newSilence(now.Add(time.Minute)),  // newer timestamp, should overwrite
+			b: state{
+				"b1": newSilence("b1", now),                   // new key, should be added
+				"a2": newSilence("a2", now.Add(-time.Minute)), // older timestamp, should be dropped
+				"a3": newSilence("a3", now.Add(time.Minute)),  // newer timestamp, should overwrite
 			},
-			final: gossipData{
-				"a1": newSilence(now),
-				"a2": newSilence(now),
-				"a3": newSilence(now.Add(time.Minute)),
-				"b1": newSilence(now),
-			},
-			delta: gossipData{
-				"b1": newSilence(now),
-				"a3": newSilence(now.Add(time.Minute)),
+			final: state{
+				"a1": newSilence("a1", now),
+				"a2": newSilence("a2", now),
+				"a3": newSilence("a3", now.Add(time.Minute)),
+				"b1": newSilence("b1", now),
 			},
 		},
 	}
 
 	for _, c := range cases {
-		ca, cb := c.a.clone(), c.b.clone()
+		for _, e := range c.b {
+			c.a.merge(e)
+		}
 
-		res := ca.Merge(cb)
-
-		require.Equal(t, c.final, res, "Merge result should match expectation")
-		require.Equal(t, c.final, ca, "Merge should apply changes to original state")
-		require.Equal(t, c.b, cb, "Merged state should remain unmodified")
-
-		ca, cb = c.a.clone(), c.b.clone()
-
-		delta := ca.mergeDelta(cb)
-
-		require.Equal(t, c.delta, delta, "Merge delta should match expectation")
-		require.Equal(t, c.final, ca, "Merge should apply changes to original state")
-		require.Equal(t, c.b, cb, "Merged state should remain unmodified")
+		require.Equal(t, c.final, c.a, "Merge result should match expectation")
 	}
 }
 
-func TestGossipDataCoding(t *testing.T) {
+func TestStateCoding(t *testing.T) {
 	// Check whether encoding and decoding the data is symmetric.
 	now := utcNow()
 
@@ -900,16 +1067,27 @@ func TestGossipDataCoding(t *testing.T) {
 
 	for _, c := range cases {
 		// Create gossip data from input.
-		in := gossipData{}
+		in := state{}
 		for _, e := range c.entries {
 			in[e.Silence.Id] = e
 		}
-		msg := in.Encode()
-		require.Equal(t, 1, len(msg), "expected single message for input")
+		msg, err := in.MarshalBinary()
+		require.NoError(t, err)
 
-		out, err := decodeGossipData(msg[0])
+		out, err := decodeState(bytes.NewReader(msg))
 		require.NoError(t, err, "decoding message failed")
 
 		require.Equal(t, in, out, "decoded data doesn't match encoded data")
 	}
 }
+
+func TestStateDecodingError(t *testing.T) {
+	// Check whether decoding copes with erroneous data.
+	s := state{"": &pb.MeshSilence{}}
+
+	msg, err := s.MarshalBinary()
+	require.NoError(t, err)
+
+	_, err = decodeState(bytes.NewReader(msg))
+	require.Equal(t, ErrInvalidState, err)
+}
